# pylint: disable=useless-import-alias
from gen_epix.commondb.policies.has_system_outage_policy import (
    HasSystemOutagePolicy as HasSystemOutagePolicy,
)
from gen_epix.commondb.policies.is_organization_admin_policy import (
    IsOrganizationAdminPolicy as IsOrganizationAdminPolicy,
)
from gen_epix.commondb.policies.read_organization_results_only_policy import (
    ReadOrganizationResultsOnlyPolicy as ReadOrganizationResultsOnlyPolicy,
)
from gen_epix.commondb.policies.read_self_results_only_policy import (
    ReadSelfResultsOnlyPolicy as ReadSelfResultsOnlyPolicy,
)
from gen_epix.commondb.policies.read_user_policy import ReadUserPolicy as ReadUserPolicy
from gen_epix.commondb.policies.update_user_policy import (
    UpdateUserPolicy as UpdateUserPolicy,
)



import datetime
from collections.abc import Callable, Iterable
from decimal import Decimal
from typing import Any
from uuid import UUID

from gen_epix.casedb.domain import command, enum, exc, model
from gen_epix.casedb.domain.policy import BaseCaseAbacPolicy
from gen_epix.casedb.services.case.base import BaseCaseService
from gen_epix.fastapp.enum import CrudOperation
from gen_epix.fastapp.unit_of_work import BaseUnitOfWork
from gen_epix.filter import UuidSetFilter
from gen_epix.filter.composite import CompositeFilter
from gen_epix.filter.string_set import StringSetFilter


def case_service_retrieve_cases_by_query(
    self: BaseCaseService, cmd: command.RetrieveCasesByQueryCommand
) -> list[UUID]:
    # TODO: This is an inefficient call first loading all cases, then filtering them and then keeping only the ids. To be replaced by optimized query.
    user, repository = self._get_user_and_repository(cmd)
    assert isinstance(user, model.User) and user.id is not None
    case_query = cmd.case_query
    case_set_ids = case_query.case_set_ids
    case_type_ids = case_query.case_type_ids
    datetime_range_filter = case_query.datetime_range_filter

    # Special case: zero case_set_ids or zero case_type_ids (None equals all)
    if case_set_ids is not None and len(case_set_ids) == 0:
        return []
    if case_type_ids is not None and len(case_type_ids) == 0:
        return []

    # @ABAC: get case abac
    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
    assert case_abac is not None
    is_full_access = case_abac.is_full_access
    has_case_read = case_abac.get_combinations_with_access_right(
        enum.CaseRight.READ_CASE
    )

    # @ABAC: Verify read access to all given case types if applicable
    if case_type_ids and not is_full_access:
        if not case_type_ids.issubset(set(has_case_read.keys())):
            raise exc.UnauthorizedAuthError(f"Unauthorized case types: {case_type_ids}")

    case_ids: list[UUID] = []
    with repository.uow() as uow:

        # @ABAC: Verify any access to all given case sets if applicable
        if case_set_ids:
            case_sets = self._retrieve_case_sets_with_content_right(
                uow,
                user.id,
                case_abac,
                # user_case_access
                enum.CaseRight.READ_CASE_SET,
            ) + self._retrieve_case_sets_with_content_right(
                uow,
                user.id,
                case_abac,
                # user_case_access
                enum.CaseRight.WRITE_CASE_SET,
            )
            invalid_case_set_ids = case_set_ids - {x.id for x in case_sets}
            if invalid_case_set_ids:
                invalid_case_set_ids_str = ", ".join(
                    [str(x) for x in invalid_case_set_ids]
                )
                raise exc.UnauthorizedAuthError(
                    f"Unauthorized case sets: {invalid_case_set_ids_str}"
                )

        # @ABAC: Verify validity of filter
        cols: list[model.Col] = []
        if case_query.filter:
            # Make sure filter keys are UUIDs
            case_query.filter.set_keys(lambda x: UUID(x) if isinstance(x, str) else x)
            cols = _verify_case_filter(self, uow, user, case_query.filter)

        # @ABAC: Retrieve all cases with read access, and content filtered on case type
        # col read access
        cases = self._retrieve_cases_with_content_right(
            uow,
            user.id,
            case_abac,
            # user_case_access,
            enum.CaseRight.READ_CASE,
            case_ids=None,
            datetime_range_filter=datetime_range_filter,
            filter_content=True,
        )

        # Filter cases by case types
        if case_type_ids:
            cases = [x for x in cases if x.case_type_id in case_type_ids]

        # Filter cases by case sets
        if case_set_ids:
            case_ids: list[UUID] = [x.id for x in cases]  # type: ignore
            case_case_sets = self._retrieve_case_case_sets_map(uow, user.id)
            cases = [
                x
                for x, y in zip(cases, case_ids)
                if y in case_case_sets and case_case_sets[y].intersection(case_set_ids)
            ]

        # Filter cases by filters
        if case_query.filter:
            map_fns = _get_map_functions_for_filters(cols)
            cases = [
                x
                for x, y in zip(
                    cases,
                    case_query.filter.match_rows(
                        (x.content for x in cases), map_fn=map_fns  # type: ignore[misc]
                    ),
                )
                if y
            ]

    # TODO: consider putting these cases, with their data already filtered, in a
    # cache, so that the expected subsequent call to retrieve them can be sped up

    # Return case ids
    case_ids: list[UUID] = [x.id for x in cases]  # type: ignore
    return case_ids


def case_service_retrieve_cases_by_id(
    self: BaseCaseService, cmd: command.RetrieveCasesByIdCommand
) -> list[model.Case]:
    case_ids = cmd.case_ids
    user, repository = self._get_user_and_repository(cmd)
    assert isinstance(user, model.User) and user.id is not None
    if not case_ids:
        return []
    # @ABAC: get case abac
    case_abac = BaseCaseAbacPolicy.get_case_abac_from_command(cmd)
    assert case_abac is not None

    with repository.uow() as uow:
        cases = self._retrieve_cases_with_content_right(
            uow,
            user.id,
            case_abac,
            enum.CaseRight.READ_CASE,
            case_ids=case_ids,
            filter_content=True,
        )
    return cases


def _verify_case_filter(
    self: BaseCaseService,
    uow: BaseUnitOfWork,
    user: model.User,
    filter: CompositeFilter,
) -> list[model.Col]:
    # Retrieve case type cols corresponding to filter keys
    filter_case_type_col_ids = filter.get_keys()
    filter_case_type_cols: list[model.CaseTypeCol] = (
        self.repository.crud(  # type:ignore[assignment]
            uow,
            user.id,
            model.CaseTypeCol,
            None,
            filter_case_type_col_ids,
            CrudOperation.READ_SOME,
        )
    )
    # Retrieve cols for case type cols
    cols: list[model.Col] = self.repository.crud(  # type:ignore[assignment]
        uow,
        user.id,
        model.Col,
        None,
        list(
            set(x.col_id for x in filter_case_type_cols)
        ),  # TODO: consider READ_SOME allowing duplicate ids
        CrudOperation.READ_SOME,
    )
    cols_ = {x.id: x for x in cols}
    cols = [cols_[x.col_id] for x in filter_case_type_cols]
    # Verify filter validity
    concept_valid_values: dict[UUID, set[str]] = {}
    region_valid_values: dict[UUID, set[str]] = {}
    for case_type_col, col, filter in zip(  # type:ignore[assignment]
        filter_case_type_cols, cols, filter.filters
    ):
        if col.concept_set_id or col.region_set_id:
            if isinstance(filter, StringSetFilter):
                valid_values = None
                if col.concept_set_id is not None:
                    # Get valid region set values
                    if col.concept_set_id not in concept_valid_values:
                        concepts: list[model.Concept] = self.app.handle(
                            command.ConceptCrudCommand(
                                user=user,
                                operation=CrudOperation.READ_ALL,
                                query_filter=UuidSetFilter(
                                    key="concept_set_id",
                                    members=frozenset({col.concept_set_id}),
                                ),
                            )
                        )
                        concept_valid_values[col.concept_set_id] = {
                            str(x.id).lower() for x in concepts
                        }
                    valid_values = concept_valid_values[col.concept_set_id]
                elif col.region_set_id is not None:
                    # Get valid region set values
                    if col.region_set_id not in region_valid_values:
                        regions: list[model.Region] = self.app.handle(
                            command.RegionCrudCommand(
                                user=user,
                                operation=CrudOperation.READ_ALL,
                                query_filter=UuidSetFilter(
                                    key="region_set_id",
                                    members=frozenset({col.region_set_id}),
                                ),
                            )
                        )
                        region_valid_values[col.region_set_id] = set(
                            [str(x.id).lower() for x in regions]
                        )
                    valid_values = region_valid_values[col.region_set_id]
                # Handle invalid values
                if valid_values is not None:
                    invalid_values = [
                        str(x)
                        for x in filter.members
                        if str(x).lower() not in valid_values
                    ]
                    if len(invalid_values):
                        invalid_values_str = ", ".join(invalid_values)
                        raise exc.InvalidArgumentsError(
                            f"Column {case_type_col.id}: invalid {filter.__class__.__name__} filter members: {invalid_values_str}"
                        )
            else:
                raise exc.InvalidArgumentsError(
                    f"Column {case_type_col.id}: invalid filter type: {filter.__class__.__name__}"
                )

    return cols


def _get_map_functions_for_filters(
    cols: Iterable[model.Col],
) -> list[Callable[[Any], Any]]:

    # Check validity of filter and generate map_fns
    map_fns = []
    for col in cols:
        if col.col_type == enum.ColType.TIME_DAY:
            map_fns.append(
                lambda x: (datetime.date.fromisoformat(x) if isinstance(x, str) else x)
            )
        elif col.col_type in {
            enum.ColType.TIME_WEEK,
            enum.ColType.TIME_MONTH,
            enum.ColType.TIME_QUARTER,
            enum.ColType.TIME_YEAR,
            enum.ColType.GEO_REGION,
            enum.ColType.NOMINAL,
            enum.ColType.ORDINAL,
            enum.ColType.INTERVAL,
            enum.ColType.TEXT,
            enum.ColType.ID_DIRECT,
            enum.ColType.ID_PSEUDONYMISED,
            enum.ColType.ORGANIZATION,
            enum.ColType.OTHER,
        }:
            map_fns.append(lambda x: x if isinstance(x, str) else str(x))
        elif col.col_type == enum.ColType.DECIMAL_0:
            map_fns.append(lambda x: int(x) if isinstance(x, str) else x)
        elif col.col_type in {
            enum.ColType.DECIMAL_1,
            enum.ColType.DECIMAL_2,
            enum.ColType.DECIMAL_3,
            enum.ColType.DECIMAL_4,
            enum.ColType.DECIMAL_5,
            enum.ColType.DECIMAL_6,
        }:
            map_fns.append(lambda x: Decimal(x) if isinstance(x, str) else x)
        elif col.col_type == enum.ColType.GEO_LATLON:
            map_fns.append(
                lambda x: (
                    (float(x.split(",")[0]), float(x.split(",")[1]))
                    if isinstance(x, str)
                    else x
                )
            )
        else:
            raise exc.InvalidArgumentsError(f"Unsupported column type: {col.col_type}")
    return map_fns



# pylint: disable=useless-import-alias
from gen_epix.commondb.domain.repository.abac import (
    BaseAbacRepository as BaseAbacRepository,
)
from gen_epix.commondb.domain.repository.organization import (
    BaseOrganizationRepository as BaseOrganizationRepository,
)
from gen_epix.commondb.domain.repository.system import (
    BaseSystemRepository as BaseSystemRepository,
)



import base64
import binascii
import re
from collections.abc import Iterable
from io import StringIO
from uuid import UUID

from Bio import SeqIO

from gen_epix.fastapp import CrudOperationSet
from gen_epix.seqdb.domain import command, enum, exc, model
from gen_epix.seqdb.domain.service.file import BaseFileService


class FileService(BaseFileService):

    VALID_NUCLEOTIDES_PATTERN = re.compile(r"^[ACGTURYSWKMBDHVN\-.]+$", re.IGNORECASE)

    def crud(  # type: ignore
        self, cmd: command.CrudCommand
    ) -> list[model.Model] | model.Model | list[UUID] | UUID:
        if isinstance(cmd, command.FileCrudCommand):
            self.validate_file(cmd)

        return super().crud(cmd)

    def validate_file(
        self,
        cmd: command.FileCrudCommand,
    ) -> None:
        is_create = cmd.operation in CrudOperationSet.CREATE.value
        if is_create:
            file_obj = cmd.get_objs()[0]
            raw_content = getattr(file_obj, "content", None)
            if (
                raw_content is None
                or (
                    isinstance(raw_content, (bytes, bytearray))
                    and len(raw_content) == 0
                )
                or (isinstance(raw_content, str) and raw_content.strip() == "")
            ):
                raise exc.InvalidArgumentsError("File content is empty")

            if isinstance(raw_content, str):
                content_bytes: bytes = raw_content.encode("utf-8")
            else:
                content_bytes = bytes(raw_content)

            # If the file doesn't start with '>' or '@' try base64-decoding it
            # (many clients embed binary files as base64 inside JSON).
            header = content_bytes.lstrip()[:1]
            if header not in (b">", b"@"):
                try:
                    decoded = base64.b64decode(content_bytes, validate=True)
                    if not decoded:
                        raise exc.InvalidArgumentsError(
                            "File content is empty after base64 decoding"
                        )
                    # replace content_bytes with decoded bytes if it looks like a FASTA/FASTQ
                    header_decoded = decoded.lstrip()[:1]
                    if header_decoded in (b">", b"@"):
                        content_bytes = decoded
                        header = header_decoded
                    else:
                        raise exc.InvalidArgumentsError(
                            "Unsupported file format: expected FASTA (lines start with >) or FASTQ (records start with @)"
                        )
                except binascii.Error as e:
                    raise exc.InvalidArgumentsError(
                        "Unsupported file format: expected FASTA (lines start with >) or FASTQ (records start with @), or a Base64-encoded representation of those."
                    ) from e

            # Determine sequence format from header byte
            match header:
                case b">":
                    sequence_format: enum.SequenceFormat = enum.SequenceFormat.FASTA
                case b"@":
                    sequence_format = enum.SequenceFormat.FASTQ
                case _:
                    raise exc.InvalidArgumentsError(
                        "Unsupported file format: expected FASTA (lines start with >) or FASTQ (records start with @)"
                    )

            try:
                text_stream: StringIO = StringIO(content_bytes.decode("utf-8"))
            except Exception as e:
                raise exc.InvalidArgumentsError(
                    "Unable to decode file content as UTF-8"
                ) from e
            try:
                sequence_records: Iterable = SeqIO.parse(text_stream, sequence_format.value.lower())  # type: ignore[no-untyped-call]
            except Exception as e:
                raise exc.InvalidArgumentsError(
                    f"Invalid {sequence_format.value} content: {e}"
                ) from e

            found_records: bool = False
            for record in sequence_records:
                found_records = True
                if sequence_format == enum.SequenceFormat.FASTQ:
                    phred_quality_scores: list[int] | None = (
                        record.letter_annotations.get("phred_quality")
                    )
                    if phred_quality_scores is None or len(phred_quality_scores) != len(
                        record.seq
                    ):
                        raise exc.InvalidArgumentsError(
                            "Invalid FASTQ: quality scores missing or length mismatch with sequence"
                        )
                sequence_string: str = str(record.seq)  # type: ignore[assignment]
                if not self.VALID_NUCLEOTIDES_PATTERN.match(sequence_string):
                    raise exc.InvalidArgumentsError(
                        f"Invalid {sequence_format.value}: sequence contains invalid characters"
                    )
                if sequence_format == enum.SequenceFormat.FASTA:
                    # SequnceFormat.FASTA doesn't need more than one record to validate
                    break

            if not found_records:
                raise exc.InvalidArgumentsError(
                    f"No sequence records found in {sequence_format.value} file"
                )



from uuid import UUID

from pydantic import Field


class DataLineageMixin:
    provenance_id: UUID | None = Field(default=None, description="Provenance ID")
    source_traceback: str | None = Field(
        default=None, description="Source traceback", max_length=255
    )



# pylint: disable=too-few-public-methods

from uuid import UUID

import sqlalchemy.orm as orm
from sqlalchemy.orm import Mapped

from gen_epix.commondb.repositories.sa_model import (
    create_mapped_column,
    create_table_args,
)
from gen_epix.seqdb.domain import DOMAIN, enum, model

Base: type = orm.declarative_base(name=enum.ServiceType.FILE.value)


class File(Base):
    __tablename__, __table_args__ = create_table_args(model.File)

    id: Mapped[UUID] = create_mapped_column(DOMAIN, model.File, "id")
    content: Mapped[bytes] = create_mapped_column(DOMAIN, model.File, "content")



from __future__ import annotations

import logging
from typing import Any
from uuid import UUID

from cachetools import TTLCache, cached

from gen_epix.commondb import policies
from gen_epix.commondb import policies as policies
from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import command, enum, exc, model, policy
from gen_epix.commondb.domain.policy import BaseReadOrganizationResultsOnlyPolicy
from gen_epix.commondb.domain.repository.abac import BaseAbacRepository
from gen_epix.commondb.domain.service import BaseAbacService
from gen_epix.fastapp import App, CrudOperation
from gen_epix.fastapp.enum import EventTiming
from gen_epix.fastapp.model import Command, CrudCommand, Policy
from gen_epix.filter import (
    CompositeFilter,
    EqualsBooleanFilter,
    EqualsUuidFilter,
    LogicalOperator,
)


class AbacService(BaseAbacService):

    CACHE_INVALIDATION_COMMANDS: tuple[type[Command], ...] = tuple()

    def __init__(
        self,
        app: App,
        repository: BaseAbacRepository,
        logger: logging.Logger | None = None,
        setup_logger: logging.Logger | None = None,
        **kwargs: Any,
    ):
        super().__init__(
            app,
            repository=repository,
            logger=logger,
            setup_logger=setup_logger,
            **kwargs,
        )
        app_impl: AppImplDetails = app.impl
        self.organization_admin_policy_model_class: type[
            model.OrganizationAdminPolicy
        ] = app_impl.get_mapped_class(model.OrganizationAdminPolicy)
        self.user_crud_command_class: type[command.UserCrudCommand] = (
            app_impl.get_mapped_class(command.UserCrudCommand)
        )
        self.is_organization_admin_policy_class: type[
            policy.BaseIsOrganizationAdminPolicy
        ] = app_impl.get_mapped_class(policies.IsOrganizationAdminPolicy)
        self.read_organization_results_only_policy_class: type[
            policy.BaseReadOrganizationResultsOnlyPolicy
        ] = app_impl.get_mapped_class(policies.ReadOrganizationResultsOnlyPolicy)
        self.read_self_results_only_policy_class: type[
            policy.BaseReadSelfResultsOnlyPolicy
        ] = app_impl.get_mapped_class(policies.ReadSelfResultsOnlyPolicy)
        self.read_user_policy_class: type[policy.BaseReadUserPolicy] = (
            app_impl.get_mapped_class(policies.ReadUserPolicy)
        )
        self.update_user_policy_class: type[policy.BaseUpdateUserPolicy] = (
            app_impl.get_mapped_class(policies.UpdateUserPolicy)
        )
        self.role_map = app_impl.role_map
        self.role_set_map = app_impl.role_set_map

    def crud(self, cmd: CrudCommand) -> Any:
        retval = super().crud(cmd)
        # Invalidate cache
        if issubclass(type(cmd), AbacService.CACHE_INVALIDATION_COMMANDS):
            self._get_user_by_id_cached.cache_clear()
        return retval

    def register_policies(
        self,
        organization_admin_write_commands: set[type[Command]] | None = None,
        read_user_commands: set[type[Command]] | None = None,
        update_user_commands: set[type[Command]] | None = None,
        read_organization_results_only_commands: set[type[Command]] | None = None,
        read_self_results_only_commands: set[type[Command]] | None = None,
    ) -> None:
        organization_admin_write_commands = (
            organization_admin_write_commands or self.ORGANIZATION_ADMIN_WRITE_COMMANDS
        )
        read_user_commands = read_user_commands or self.READ_USER_COMMANDS
        update_user_commands = update_user_commands or self.UPDATE_USER_COMMANDS
        read_organization_results_only_commands = (
            read_organization_results_only_commands
            or self.READ_ORGANIZATION_RESULTS_ONLY_COMMANDS
        )
        read_self_results_only_commands = (
            read_self_results_only_commands or self.READ_SELF_RESULTS_ONLY_COMMANDS
        )
        f = self.app.register_policy
        policy: Policy
        command_class: type[Command]
        policy = self.is_organization_admin_policy_class(self)
        for command_class in organization_admin_write_commands:
            f(command_class, policy, EventTiming.BEFORE)
        policy = self.read_user_policy_class(self)
        for command_class in read_user_commands:
            f(command_class, policy, EventTiming.AFTER)
        policy = self.update_user_policy_class(self)
        for command_class in update_user_commands:
            f(command_class, policy, EventTiming.BEFORE)
        policy = self.read_organization_results_only_policy_class(self)
        for command_class in read_organization_results_only_commands:
            f(command_class, policy, EventTiming.DURING)
            f(command_class, policy, EventTiming.AFTER)
        policy = self.read_self_results_only_policy_class(self)
        for command_class in read_self_results_only_commands:
            f(command_class, policy, EventTiming.AFTER)

    def retrieve_organizations_under_admin(
        self, cmd: command.RetrieveOrganizationsUnderAdminCommand
    ) -> set[UUID]:
        assert cmd.user and cmd.user.id
        # Special case: user has a role that makes them admin of all organizations
        is_all_organizations = False
        for policy in cmd._policies:
            if isinstance(policy, BaseReadOrganizationResultsOnlyPolicy):
                is_all_organizations = (
                    len(
                        cmd.user.roles.intersection(
                            policy.role_set_map[enum.RoleSet.GE_APP_ADMIN]
                        )
                    )
                    > 0
                )
                break
        if is_all_organizations:
            organizations: list[model.Organization] = self.app.handle(
                command.OrganizationCrudCommand(
                    user=cmd.user,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                )
            )
            return {x.id for x in organizations}  # type: ignore[misc]
        # Retrieve organizations for which the user is an admin
        with self.repository.uow() as uow:
            organization_admin_policies: list[model.OrganizationAdminPolicy] = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    user_id=cmd.user.id,
                    model_class=self.organization_admin_policy_model_class,
                    objs=None,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    filter=CompositeFilter(
                        operator=LogicalOperator.AND,
                        filters=[
                            EqualsUuidFilter(key="user_id", value=cmd.user.id),
                            EqualsBooleanFilter(key="is_active", value=True),
                        ],
                    ),
                )
            )
        return set(x.organization_id for x in organization_admin_policies)

    def retrieve_organization_admin_name_emails(
        self,
        cmd: command.RetrieveOrganizationAdminNameEmailsCommand,
    ) -> list[model.UserNameEmail]:
        if not isinstance(cmd.user, model.User):
            raise exc.ServiceException(
                "Command has no or wrong user type: {cmd.user.__class__.__name__}"
            )
        with self.repository.uow() as uow:
            organization_admin_policies: list[model.OrganizationAdminPolicy] = (
                self.repository.crud(  # type:ignore[assignment]
                    uow,
                    user_id=cmd.user.id,
                    model_class=self.organization_admin_policy_model_class,
                    objs=None,
                    obj_ids=None,
                    operation=CrudOperation.READ_ALL,
                    filter=EqualsUuidFilter(
                        key="organization_id", value=cmd.user.organization_id
                    ),
                )
            )
        organization_admin_user_ids = {
            x.user_id
            for x in organization_admin_policies
            if x.organization_id == cmd.user.organization_id and x.is_active
        }
        users = self.app.handle(
            self.user_crud_command_class(
                user=cmd.user,
                obj_ids=list(organization_admin_user_ids),
                operation=CrudOperation.READ_SOME,
            )
        )
        return [
            model.UserNameEmail(
                id=x.id,
                name=x.name,
                email=x.email,
            )
            for x in users
        ]

    def temp_update_user_own_organization(
        self,
        cmd: command.UpdateUserOwnOrganizationCommand,
    ) -> model.User:
        """
        Behaviour:
        - Update User.organization
        - Delete any OrganizationAdminPolicy for the user and their previous
          organization
        """
        is_new_user = cmd.is_new_user
        tgt_organization_id = cmd.organization_id
        assert cmd.user is not None
        user: model.User = cmd.user
        assert user.id is not None

        # Special case: new organization is same as current
        if user.organization_id == tgt_organization_id and not is_new_user:
            return user

        with self.repository.uow() as uow:
            # Delete any OrganizationAdminPolicy for the user and their previous organization
            if not is_new_user:
                self.repository.crud(
                    uow,
                    user.id,
                    self.organization_admin_policy_model_class,
                    None,
                    None,
                    CrudOperation.DELETE_ALL,
                    filter=CompositeFilter(
                        operator=LogicalOperator.AND,
                        filters=[
                            EqualsUuidFilter(key="user_id", value=user.id),
                            EqualsUuidFilter(
                                key="organization_id", value=user.organization_id
                            ),
                        ],
                    ),
                )
            # Change the user organization
            updated_user = user.model_copy()
            updated_user.organization_id = tgt_organization_id
            user = self.app.handle(
                self.user_crud_command_class(
                    user=user,
                    objs=updated_user,
                    operation=CrudOperation.UPDATE_ONE,
                )
            )

        # Invalidate cache
        # TODO: develop general system for caching and cache invalidation
        self._get_user_by_id_cached.cache_clear()
        return user

    @cached(cache=TTLCache(maxsize=1024, ttl=300))
    def _get_user_by_id_cached(self, user_id: UUID) -> model.User:
        user: model.User = self.app.handle(
            self.user_crud_command_class(
                user=None,
                obj_ids=user_id,
                operation=CrudOperation.READ_ONE,
            )
        )
        return user



# pylint: disable=useless-import-alias
from gen_epix.fastapp.services.rbac.policy import RbacPolicy as RbacPolicy
from gen_epix.fastapp.services.rbac.service import BaseRbacService as BaseRbacService



# pylint: disable=useless-import-alias
from gen_epix.fastapp.domain.domain import Domain as Domain
from gen_epix.fastapp.domain.entity import Entity as Entity
from gen_epix.fastapp.domain.key import Key as Key
from gen_epix.fastapp.domain.link import Link as Link
from gen_epix.fastapp.domain.util import create_keys as create_keys
from gen_epix.fastapp.domain.util import create_links as create_links



# pylint: disable=useless-import-alias

from gen_epix.commondb.domain.service import (
    BaseOrganizationService as BaseOrganizationService,
)
from gen_epix.commondb.domain.service import BaseRbacService as BaseRbacService
from gen_epix.commondb.domain.service import BaseSystemService as BaseSystemService
from gen_epix.fastapp.services.auth import BaseAuthService as BaseAuthService
from gen_epix.seqdb.domain.service.abac import BaseAbacService as BaseAbacService
from gen_epix.seqdb.domain.service.seq import BaseSeqService as BaseSeqService



from decimal import Decimal
from typing import Literal

from pydantic import Field

from gen_epix.filter.enum import FilterType
from gen_epix.filter.equals import EqualsFilter


class EqualsNumberFilter(EqualsFilter):
    value: int | float | Decimal = Field(
        description="The number to match.", frozen=True
    )


class TypedEqualsNumberFilter(EqualsNumberFilter):
    type: Literal[FilterType.EQUALS_NUMBER.value]



from collections.abc import Callable, Hashable, Iterable, Iterator
from typing import Any, Literal

from pydantic import BaseModel

from gen_epix.filter.base import Filter
from gen_epix.filter.enum import FilterType


class NoFilter(Filter):
    key: Literal[False] = False

    def _match(self, value: Any) -> bool:
        return not self.invert

    def match_column(
        self,
        values: Iterable[Any | None],
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
    ) -> Iterator[bool]:
        for value in values:
            yield not self.invert

    def filter_column(
        self,
        values: Iterable[Any | None],
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
    ) -> Iterator[Any | None]:
        for value in values:
            if not self.invert:
                yield value

    def match_row(
        self,
        row: dict[Hashable, Any | None] | BaseModel,
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
        is_model: bool = False,
    ) -> bool:
        return not self.invert

    def match_rows(
        self,
        rows: Iterable[dict[Hashable, Any | None] | BaseModel],
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
        is_model: bool = False,
    ) -> Iterator[bool]:
        for row in rows:
            if not self.invert:
                yield True

    def filter_rows(
        self,
        rows: Iterable[dict[Hashable, Any | None] | BaseModel],
        na_values: set[Any] | None = None,
        map_fn: Callable[[Any], Any] | None = None,
        is_model: bool = False,
    ) -> Iterator[dict[Hashable, Any | None]]:
        for row in rows:
            if not self.invert:
                yield row


class TypedNoFilter(NoFilter):
    type: Literal[FilterType.NO_FILTER.value]



from collections.abc import Callable
from typing import Any, NoReturn

from fastapi import APIRouter, FastAPI

from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import command, enum, model
from gen_epix.fastapp import App
from gen_epix.fastapp.api import CrudEndpointGenerator


def create_auth_endpoints(
    router: APIRouter | FastAPI,
    app: App,
    service_type: enum.ServiceType = enum.ServiceType.AUTH,
    handle_exception: Callable[[str, Any, Exception], NoReturn] | None = None,
    **kwargs: Any,
) -> None:
    assert handle_exception
    app_impl: AppImplDetails = app.impl
    registered_user_dependency = app_impl.registered_user_dependency

    # Specific endpoints - Auth
    @router.get(
        "/identity_providers",
        operation_id="identity_providers__get_all",
        name="IdentityProvider",
        description="Get all public identity providers",
    )
    async def identity_providers__get_all() -> list[model.IdentityProvider]:
        try:
            cmd = command.GetIdentityProvidersCommand(user=None, public=True)
            retval: list[model.IdentityProvider] = app.handle(cmd)
        except Exception as exception:
            handle_exception("3ddf8ebb", None, exception)
        return retval

    # CRUD
    crud_endpoint_sets = CrudEndpointGenerator.create_crud_endpoint_set_for_domain(
        app,
        service_type=service_type,
        user_dependency=registered_user_dependency,
    )
    CrudEndpointGenerator.generate_endpoints(
        router, crud_endpoint_sets, handle_exception
    )



from typing import Any

from gen_epix.fastapp.services.auth.literal import EMAIL_PATTERN


def get_email_from_claims(
    claims: dict[str, Any],
) -> str | None:
    email = claims.get("email")
    if email is None:
        for claim in claims.values():
            if isinstance(claim, str) and EMAIL_PATTERN.match(claim.lower()):
                return claim.lower()
    else:
        return email.lower()
    return None


def get_name_from_claims(
    claims: dict[str, Any], name_claims: list[str | list[str]]
) -> str | None:
    """
    Get the name from the claims, checking against a list of possible name claims.
    """
    for name_claim in name_claims:
        if isinstance(name_claim, str):
            if name_claim in claims:
                return str(claims[name_claim])
        else:
            # Check if every subclaim exists and if so return space concatenated string
            values = [claims[x] for x in name_claim if x in claims]
            if len(values) == len(name_claim):
                return " ".join(str(x) for x in values)
    return None



from collections.abc import Callable
from enum import Enum
from typing import Any, NoReturn
from uuid import UUID

from fastapi import APIRouter, FastAPI
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.domain import DOMAIN, command, enum, model
from gen_epix.commondb.util import copy_model_field
from gen_epix.fastapp import App
from gen_epix.fastapp.api.crud_endpoint_generator import CrudEndpointGenerator
from gen_epix.fastapp.enum import PermissionType
from gen_epix.fastapp.model import Permission

CommandName = Enum("CommandName", {x: x for x in DOMAIN.command_names})  # type: ignore[misc] # Dynamic Enum required


class ApiPermission(PydanticBaseModel, frozen=True):
    command_name: CommandName = (  # pyright: ignore[reportInvalidTypeForm]
        copy_model_field(Permission, "command_name")
    )
    permission_type: PermissionType = copy_model_field(Permission, "permission_type")


class UserInvitationRequestBody(PydanticBaseModel):
    key: str = copy_model_field(model.UserInvitation, "key")
    roles: set[str] = copy_model_field(model.UserInvitation, "roles")
    organization_id: UUID = copy_model_field(model.UserInvitation, "organization_id")


class UpdateOrganizationSetOrganizationRequestBody(PydanticBaseModel):
    organization_set_members: list[model.OrganizationSetMember] = Field(
        description="The updated set of organization set members, replacing the previous set"
    )


class UpdateDataCollectionSetDataCollectionRequestBody(PydanticBaseModel):
    data_collection_set_members: list[model.DataCollectionSetMember] = Field(
        description="The updated set of data collection set members, replacing the previous set"
    )


class UpdateUserRequestBody(PydanticBaseModel):
    is_active: bool | None = Field(
        description="The updated active status of the user. Not updated if not provided."
    )
    roles: set[str] | None = Field(
        description="The updated set of roles of the user. Not updated if not provided. If provided, should have at least one element.",
    )
    organization_id: UUID | None = Field(
        description="The updated organization ID of the user. Not updated if not provided."
    )


class UpdateUserOwnOrganizationRequestBody(PydanticBaseModel):
    organization_id: UUID = Field(
        description="The ID of the organization to update the user to"
    )


def create_organization_endpoints(
    router: APIRouter | FastAPI,
    app: App,
    service_type: enum.ServiceType = enum.ServiceType.ORGANIZATION,
    handle_exception: Callable[[str, Any, Exception], NoReturn] | None = None,
    api_permission_class: type = Permission,
    **kwargs: Any,
) -> None:
    assert handle_exception
    app_impl: AppImplDetails = app.impl
    user_class: type[model.User] = app_impl.get_mapped_class(model.User)
    user_invitation_class: type[model.UserInvitation] = app_impl.get_mapped_class(
        model.UserInvitation
    )
    invite_user_command_class: type[command.InviteUserCommand] = (
        app_impl.get_mapped_class(command.InviteUserCommand)
    )
    retrieve_invite_user_constraints_command_class: type[
        command.RetrieveInviteUserConstraintsCommand
    ] = app_impl.get_mapped_class(command.RetrieveInviteUserConstraintsCommand)
    update_user_command_class: type[command.UpdateUserCommand] = (
        app_impl.get_mapped_class(command.UpdateUserCommand)
    )
    registered_user_dependency = app_impl.registered_user_dependency
    new_user_dependency = app_impl.new_user_dependency

    @router.post(
        "/invite_user",
        operation_id="invite_user",
        name="Invite a user",
        description=invite_user_command_class.__doc__,
    )
    async def invite_user(
        user: registered_user_dependency, user_invitation: UserInvitationRequestBody  # type: ignore[valid-type] # Dynamic type annotation
    ) -> user_invitation_class:  # type: ignore
        try:
            retval: user_invitation_class = app.handle(  # type: ignore[valid-type] # Dynamic type annotation
                invite_user_command_class(
                    user=user,
                    key=user_invitation.key,
                    roles=user_invitation.roles,
                    organization_id=user_invitation.organization_id,
                )
            )
        except Exception as exception:
            handle_exception("e088de91", None, exception)
        return retval

    @router.get(
        "/invite_user/constraints",
        operation_id="invite_user__constraints",
        name="The constraints for inviting a user",
        description=retrieve_invite_user_constraints_command_class.__doc__,
    )
    async def invite_user__constraints(
        user: registered_user_dependency,
    ) -> model.UserInvitationConstraints:
        try:
            retval: model.UserInvitationConstraints = app.handle(  # type: ignore[valid-type]
                retrieve_invite_user_constraints_command_class(user=user)
            )
        except Exception as exception:
            handle_exception("cad2509e", None, exception)
        return retval

    @router.post(
        "/user_registrations/{token}",
        operation_id="user_registrations__post_one",
        name="RegisterInvitedUser",
        description=command.RegisterInvitedUserCommand.__doc__,
    )
    async def user_registrations__post_one(
        user: new_user_dependency, token: str  # type: ignore[valid-type] # Dynamic type annotation
    ) -> user_class:  # type: ignore[valid-type] # Dynamic type annotation
        try:
            cmd = command.RegisterInvitedUserCommand(
                user=user,
                token=token,
            )
            retval: user_class = app.handle(cmd)  # type: ignore[valid-type] # Dynamic type annotation
        except Exception as exception:
            handle_exception("fc1fc53c", None, exception)
        return retval

    @router.put(
        "/organization_sets/{organization_set_id}/organizations",
        operation_id="organization_sets__put__organizations",
        name="OrganizationSet_Organization",
        description=command.OrganizationSetOrganizationUpdateAssociationCommand.__doc__,
    )
    async def organization_sets__put__organizations(
        user: registered_user_dependency,  # type: ignore
        organization_set_id: UUID,
        request_body: UpdateOrganizationSetOrganizationRequestBody,
    ) -> list[model.OrganizationSetMember]:
        try:
            cmd = command.OrganizationSetOrganizationUpdateAssociationCommand(
                user=user,
                obj_id1=organization_set_id,
                association_objs=request_body.organization_set_members,
                props={"return_id": False},
            )
            retval: list[model.OrganizationSetMember] = app.handle(cmd)
        except Exception as exception:
            handle_exception("c026628e", user, exception)
        return retval

    @router.put(
        "/data_collection_sets/{data_collection_set_id}/data_collections",
        operation_id="data_collection_sets__put__data_collections",
        name="DataCollectionSet_DataCollection",
        description=command.DataCollectionSetDataCollectionUpdateAssociationCommand.__doc__,
    )
    async def data_collection_sets__put__data_collections(
        user: registered_user_dependency,  # type: ignore
        data_collection_set_id: UUID,
        request_body: UpdateDataCollectionSetDataCollectionRequestBody,
    ) -> list[model.DataCollectionSetMember]:
        try:
            cmd = command.DataCollectionSetDataCollectionUpdateAssociationCommand(
                user=user,
                obj_id1=data_collection_set_id,
                association_objs=request_body.data_collection_set_members,
                props={"return_id": False},
            )
            retval: list[model.DataCollectionSetMember] = app.handle(cmd)
        except Exception as exception:
            handle_exception("cf892de0", user, exception)
        return retval

    @router.get(
        "/user_me",
        operation_id="user_me__get_one",
        name="UserMe",
        description=user_class.__doc__,
    )
    async def user_me__get_one(
        user: registered_user_dependency,  # type: ignore
    ) -> user_class:
        return user

    @router.get(
        "/user_me/permissions",
        operation_id="user_me__retrieve_permissions",
        name="UserMe_Permissions",
        description=command.RetrieveOwnPermissionsCommand.__doc__,
    )
    async def user_me__retrieve_permissions(
        user: registered_user_dependency,  # type: ignore
    ) -> set[api_permission_class]:  # pyricht: ignore[reportInvalidTypeForm]
        try:
            cmd = command.RetrieveOwnPermissionsCommand(user=user)
            permissions: set[Permission] = app.handle(cmd)
            retval = {api_permission_class(**x.model_dump()) for x in permissions}
        except Exception as exception:
            handle_exception("a7f3b8e2", user, exception)
        return retval

    @router.put(
        "/users/{object_id}",
        operation_id="users__put_one",
        name="UpdateUser",
        description=update_user_command_class.__doc__,
    )
    async def users__put_one(
        user: registered_user_dependency, object_id: UUID, request_body: UpdateUserRequestBody  # type: ignore
    ) -> user_class:
        try:
            cmd = update_user_command_class(
                user=user,
                tgt_user_id=object_id,
                is_active=request_body.is_active,
                roles=request_body.roles,
                organization_id=request_body.organization_id,
            )
            retval: user_class = app.handle(cmd)
        except Exception as exception:
            handle_exception("a594ba2b", None, exception)
        return retval

    @router.put(
        "/update_user_own_organization",
        operation_id="update_user_own_organization",
        name="UpdateUserOwnOrganizationCommand",
        description=command.UpdateUserOwnOrganizationCommand.__doc__,
    )
    async def update_user_own_organization(
        user: registered_user_dependency, data: UpdateUserOwnOrganizationRequestBody  # type: ignore
    ) -> user_class:
        try:
            cmd = command.UpdateUserOwnOrganizationCommand(
                user=user,
                organization_id=data.organization_id,
            )
            retval: model.User = app.handle(cmd)
        except Exception as exception:
            handle_exception("c2382b65", None, exception)
        return retval

    # CRUD
    crud_endpoint_sets = CrudEndpointGenerator.create_crud_endpoint_set_for_domain(
        app,
        service_type=service_type,
        user_dependency=registered_user_dependency,
    )
    CrudEndpointGenerator.generate_endpoints(
        router, crud_endpoint_sets, handle_exception
    )



import json
import ssl
from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import Any
from uuid import UUID

import httpx
from pydantic import BaseModel as PydanticBaseModel

from gen_epix.fastapp import exc, model
from gen_epix.fastapp.api.crud_endpoint_generator import CrudEndpointGenerator
from gen_epix.fastapp.app import App
from gen_epix.fastapp.domain.domain import Domain
from gen_epix.fastapp.enum import CrudOperation, EventTiming, HttpProtocol, StringCasing
from gen_epix.fastapp.exc import ServiceException
from gen_epix.fastapp.model import Command, CrudCommand, Policy
from gen_epix.fastapp.util import create_ssl_context


class RemoteApp(App):

    DEFAULT_ROUTE_PREFIX = "/"

    DEFAULT_REQUEST_HEADERS: dict[str, str] = {"Content-Type": "application/json"}

    def __init__(
        self,
        domain: Domain,
        host: str,
        port: int | None,
        http_protocol: HttpProtocol = HttpProtocol.HTTPS,
        default_route_prefix: str | None = None,
        default_headers: dict[str, str] | None = None,
        add_generated_crud_route_handlers: bool = True,
        ssl_cert_file: Path | str | None = None,
        disable_ssl_verification: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(domain, **kwargs)
        self._host = host
        self._port = port
        self._http_protocol = http_protocol
        self._default_route_prefix = default_route_prefix or self.DEFAULT_ROUTE_PREFIX
        self._default_headers = default_headers or self.DEFAULT_REQUEST_HEADERS
        self._routes: dict[type[Command], str] = {}

        # Initialise SSL context
        self._ssl_context = create_ssl_context(
            host, ssl_cert_file, disable_ssl_verification
        )

        # Create and register generated crud route handlers
        if add_generated_crud_route_handlers:
            for command_class in self.domain.crud_commands:
                base_route = self.register_generated_crud_route(command_class)
                handler = self.create_generated_crud_route_handler(
                    command_class,
                    base_route,
                )
                self.register_handler(command_class, handler)

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int | None:
        return self._port

    @property
    def http_protocol(self) -> HttpProtocol:
        return self._http_protocol

    @property
    def host_url(self) -> str:
        port_str = f":{self._port}" if self._port else ""
        return f"{self.http_protocol.value.lower()}://{self._host}{port_str}"

    @property
    def ssl_context(self) -> ssl.SSLContext | bool:
        return self._ssl_context

    def register_policy(
        self,
        command_class: type[Command],
        policy: Policy,
        timing: EventTiming = EventTiming.BEFORE,
    ) -> None:
        raise ServiceException("Policies cannot be registered on RemoteApp instances")

    def unregister_policy(
        self, command_class: type[Command], policy: Policy, timing: EventTiming
    ) -> None:
        raise ServiceException("Policies cannot be unregistered on RemoteApp instances")

    def register_route(
        self,
        command_class: type[Command],
        route: str,
        add_host: bool = True,
        add_prefix: bool = True,
    ) -> str:
        """
        Registers the route that is able to handle the command after it is
        converted into a request by the handler.
        """
        if command_class in self._routes:
            raise ServiceException(
                f"Route already registered for command: {command_class.__name__}"
            )
        route = self._default_route_prefix + route if add_prefix else route
        route = route if not add_host else f"{self.host_url}{route}"
        self._routes[command_class] = route
        return route

    def unregister_route(self, command_class: type[Command]) -> None:
        if command_class not in self._routes:
            raise ServiceException(
                f"No route registered for command: {command_class.__name__}"
            )
        del self._routes[command_class]

    def get_route(self, cmd: Command) -> str:
        route = self._routes.get(cmd.__class__, None)
        if not route:
            raise NotImplementedError(
                f"No route registered for command: {cmd.__class__.__name__}"
            )
        return route

    def get_headers(self, cmd: Command) -> dict[str, str]:
        """
        Get headers for the command. Override to include e.g. authorization header.
        """
        return self._default_headers

    def apply_handler(
        self,
        cmd: Command,
        handler: Callable[[Command], Any],
    ) -> Any:
        command_class = cmd.__class__
        route = self._routes.get(command_class, None)
        if not route:
            raise NotImplementedError(
                f"No route registered for command: {cmd.__class__.NAME}"
            )
        try:
            retval = handler(cmd)
        except httpx.RequestError as e:
            raise exc.ServiceException(
                f"HTTP request error when handling remote command {command_class.NAME}: {e}"
            ) from e
        except httpx.HTTPStatusError as e:
            # Handle HTTPStatusError with proper access to response attributes
            status_code = (
                getattr(e.response, "status_code", "unknown")
                if hasattr(e, "response") and e.response
                else "unknown"
            )
            raise exc.ServiceException(
                f"HTTP status {status_code} error when handling remote command {command_class.NAME}: {e}"
            ) from e
        except Exception as e:
            raise exc.ServiceException(
                f"Error when handling remote command {command_class.NAME}: {e}"
            ) from e
        return retval

    def register_generated_crud_route(
        self,
        command_class: type[CrudCommand],
        route_root: str | None = None,
        add_host: bool = True,
        add_prefix: bool = True,
        route_root_casing: StringCasing = StringCasing.SNAKE_CASE,
        route_root_plural: bool = True,
    ) -> str:
        model_class = command_class.MODEL_CLASS
        entity = model_class.ENTITY
        assert entity is not None
        route_root = route_root or entity.get_name_by_casing(
            route_root_casing, is_plural=route_root_plural
        )
        assert route_root is not None
        return self.register_route(
            command_class, route_root, add_host=add_host, add_prefix=add_prefix
        )

    def create_generated_crud_route_handler(
        self,
        command_class: type[CrudCommand],
        base_route: str,
        batch_route_suffix: str | None = None,
        query_route_suffix: str | None = None,
        ids_route_suffix: str | None = None,
    ) -> Callable[[Command], Any]:
        batch_route_suffix = (
            batch_route_suffix or CrudEndpointGenerator.DEFAULT_BATCH_ROUTE_SUFFIX
        )
        query_route_suffix = (
            query_route_suffix or CrudEndpointGenerator.DEFAULT_QUERY_ROUTE_SUFFIX
        )
        ids_route_suffix = (
            ids_route_suffix or CrudEndpointGenerator.DEFAULT_IDS_ROUTE_SUFFIX
        )
        model_class = command_class.MODEL_CLASS
        entity = model_class.ENTITY
        assert entity is not None

        def handler(
            base_route: str,
            batch_route_suffix: str,
            query_route_suffix: str,
            ids_route_suffix: str,
            cmd: CrudCommand,
        ) -> Any:

            headers = self.get_headers(cmd)
            model_class = cmd.MODEL_CLASS
            return_model_class: type = model_class
            is_list = False
            with httpx.Client(verify=self.ssl_context) as client:
                if cmd.operation == CrudOperation.READ_ALL:
                    if cmd.query_filter:
                        response = client.post(
                            base_route
                            + query_route_suffix
                            + (
                                "/" + ids_route_suffix
                                if cmd.props.get("return_id", False)
                                else ""
                            ),
                            json=json.loads(cmd.query_filter.model_dump_json()),
                            headers=headers,
                        )
                    else:
                        response = client.get(base_route, headers=headers)
                    is_list = True
                elif cmd.operation == CrudOperation.READ_SOME:
                    assert isinstance(cmd.obj_ids, list)
                    ids = json.dumps([str(x) for x in cmd.obj_ids])
                    response = client.get(
                        base_route + batch_route_suffix,
                        headers=headers,
                        params={"ids": ids},
                    )
                    is_list = True
                elif cmd.operation == CrudOperation.READ_ONE:
                    response = client.get(
                        f"{base_route}/{cmd.obj_ids}",
                        headers=headers,
                    )
                elif cmd.operation == CrudOperation.CREATE_ONE:
                    assert isinstance(cmd.objs, model.Model)
                    response = client.post(
                        f"{base_route}",
                        json=json.loads(cmd.objs.model_dump_json()),
                        headers=headers,
                    )
                elif cmd.operation == CrudOperation.CREATE_SOME:
                    assert isinstance(cmd.objs, list)
                    response = client.post(
                        base_route + batch_route_suffix,
                        json=[json.loads(x.model_dump_json()) for x in cmd.objs],
                        headers=headers,
                    )
                    is_list = True
                elif cmd.operation == CrudOperation.UPDATE_ONE:
                    assert isinstance(cmd.objs, model.Model)
                    response = client.put(
                        f"{base_route}/{cmd.objs.id}",  # type: ignore[attr-defined]
                        json=json.loads(cmd.objs.model_dump_json()),
                        headers=headers,
                    )
                elif cmd.operation == CrudOperation.UPDATE_SOME:
                    assert isinstance(cmd.objs, list)
                    response = client.put(
                        f"{base_route}",
                        json=[json.loads(x.model_dump_json()) for x in cmd.objs],
                        headers=headers,
                    )
                    is_list = True
                elif cmd.operation == CrudOperation.DELETE_ONE:
                    assert isinstance(cmd.obj_ids, UUID)
                    response = client.delete(
                        f"{base_route}/{cmd.obj_ids}", headers=headers
                    )
                    return_model_class = UUID
                elif cmd.operation == CrudOperation.DELETE_SOME:
                    assert isinstance(cmd.obj_ids, list)
                    ids = json.dumps([str(x) for x in cmd.obj_ids])
                    response = client.delete(
                        base_route + batch_route_suffix,
                        headers=headers,
                        params={"ids": ids},
                    )
                    return_model_class = UUID
                    is_list = True
                else:
                    raise NotImplementedError(f"Unsupported operation: {cmd.operation}")
                response.raise_for_status()
            retval = self._content_to_obj(response, return_model_class, is_list=is_list)
            return retval

        return partial(
            handler,
            base_route,
            batch_route_suffix,
            query_route_suffix,
            ids_route_suffix,
        )

    @staticmethod
    def _content_to_obj(
        response: httpx.Response, retval_class: type, is_list: bool = False
    ) -> Any:
        if response.status_code not in (200, 201):
            return None
        decoded_obj = json.loads(response.content.decode(response.encoding or "utf-8"))
        if issubclass(retval_class, PydanticBaseModel):
            if is_list:
                return [retval_class(**x) for x in decoded_obj]
            else:
                return retval_class(**decoded_obj)
        elif issubclass(retval_class, UUID):
            if is_list:
                return [UUID(x) for x in decoded_obj]
            else:
                return UUID(decoded_obj)
        raise NotImplementedError(f"Unsupported return type: {retval_class}")



import json
import uuid
from collections.abc import Callable
from datetime import datetime, timedelta
from typing import Any
from uuid import UUID

# import libraries
from fastapi import FastAPI
from fastapi.testclient import TestClient
from httpx import Response
import jwt
from pydantic import BaseModel as PydanticBaseModel

from gen_epix.commondb.api import UpdateUserRequestBody, UserInvitationRequestBody
from gen_epix.commondb.domain import command, model
from gen_epix.fastapp import App, Command, CrudCommand, CrudOperation


class EndpointTestClient:

    SECRET_KEY = str(uuid.uuid4())
    ENCRYPTION_ALGORITHM = "HS256"

    def __init__(
        self,
        app: App,
        fast_api: FastAPI,
        app_last_handled_exception: dict,
        user_class: type[model.User] = model.User,
        user_invitation_class: type[model.UserInvitation] = model.UserInvitation,
        user_invitation_constraints_class: type[
            model.UserInvitationConstraints
        ] = model.UserInvitationConstraints,
        organization_admin_policy_class: type[
            model.OrganizationAdminPolicy
        ] = model.OrganizationAdminPolicy,
        user_crud_command_class: type[
            command.UserCrudCommand
        ] = command.UserCrudCommand,
        user_invitation_crud_command_class: type[
            command.UserInvitationCrudCommand
        ] = command.UserInvitationCrudCommand,
        organization_admin_policy_crud_command_class: type[
            command.OrganizationAdminPolicyCrudCommand
        ] = command.OrganizationAdminPolicyCrudCommand,
        retrieve_invite_user_constraints_command_class: type[
            command.RetrieveInviteUserConstraintsCommand
        ] = command.RetrieveInviteUserConstraintsCommand,
        invite_user_command_class: type[
            command.InviteUserCommand
        ] = command.InviteUserCommand,
        register_invited_user_command_class: type[
            command.RegisterInvitedUserCommand
        ] = command.RegisterInvitedUserCommand,
        retrieve_organization_admin_name_emails_command_class: type[
            command.RetrieveOrganizationAdminNameEmailsCommand
        ] = command.RetrieveOrganizationAdminNameEmailsCommand,
        update_user_command_class: type[
            command.UpdateUserCommand
        ] = command.UpdateUserCommand,
        user_invitation_request_body: type[
            PydanticBaseModel
        ] = UserInvitationRequestBody,
        update_user_request_body: type[PydanticBaseModel] = UpdateUserRequestBody,
        register_crud_commands: bool = True,
        route_prefix: str | None = None,
    ):
        self.app = app
        self.fast_api = fast_api
        self.test_client = TestClient(fast_api, raise_server_exceptions=False)
        self.user_class = user_class
        self.user_invitation_class = user_invitation_class
        self.user_invitation_constraints_class = user_invitation_constraints_class
        self.organization_admin_policy_class = organization_admin_policy_class
        self.user_crud_command_class = user_crud_command_class
        self.user_invitation_crud_command_class = user_invitation_crud_command_class
        self.organization_admin_policy_crud_command_class = (
            organization_admin_policy_crud_command_class
        )
        self.retrieve_invite_user_constraints_command_class = (
            retrieve_invite_user_constraints_command_class
        )
        self.invite_user_command_class = invite_user_command_class
        self.register_invited_user_command_class = register_invited_user_command_class
        self.retrieve_organization_admin_name_emails_command_class = (
            retrieve_organization_admin_name_emails_command_class
        )
        self.update_user_command_class = update_user_command_class
        self.user_invitation_request_body = user_invitation_request_body
        self.update_user_request_body = update_user_request_body
        self.route_prefix = route_prefix or ""
        self._handlers: dict[
            type[Command],
            Callable[[Command, str, dict[str, str] | None], tuple[Any, Response]],
        ] = {}
        if register_crud_commands:
            for crud_command_class in app.domain.crud_commands:
                self.register_handler(crud_command_class, self.handle_crud_command)  # type: ignore[arg-type]
        self.register_handler(
            command.GetIdentityProvidersCommand, self.handle_get_identity_providers
        )
        self.register_handler(self.invite_user_command_class, self.handle_invite_user)
        self.register_handler(
            self.retrieve_invite_user_constraints_command_class,
            self.handle_retrieve_invite_user_constraints,
        )
        self.register_handler(
            self.register_invited_user_command_class, self.handle_register_invited_user
        )
        self.register_handler(self.update_user_command_class, self.handle_update_user)

    def register_handler(
        self,
        command_class: type[Command],
        handler: Callable[[Command, str, dict[str, str] | None], tuple[Any, Response]],
    ) -> None:
        self._handlers[command_class] = handler

    def handle(
        self,
        cmd: Command,
        return_response: bool = False,
        route_prefix: str | None = None,
        **kwargs: Any,
    ) -> Any:
        route_prefix = route_prefix or self.route_prefix
        if cmd.user:
            headers = self.get_headers(cmd)
        else:
            headers = None
        handler = self._handlers.get(cmd.__class__)
        if not handler:
            raise NotImplementedError(f"Unsupported command: {cmd.__class__.__name__}")
        retval, response = handler(cmd, route_prefix, headers)
        if return_response:
            return retval, response
        return retval

    def handle_get_identity_providers(
        self,
        cmd: command.GetIdentityProvidersCommand,
        route_prefix: str,
        headers: dict[str, str] | None,
    ) -> tuple[Any, Response]:
        response = self.test_client.get(route_prefix + "/identity_providers")
        retval = self._content_to_obj(response, model.IdentityProvider, is_list=True)
        return retval, response

    def handle_invite_user(
        self,
        cmd: command.InviteUserCommand,
        route_prefix: str,
        headers: dict[str, str] | None,
    ) -> tuple[Any, Response]:
        request_body = self.user_invitation_request_body(
            key=cmd.key,
            email=cmd.email,
            name=cmd.name,
            roles=cmd.roles,
            organization_id=cmd.organization_id,
        )
        response = self.test_client.post(
            route_prefix + "/invite_user",
            json=json.loads(request_body.model_dump_json()),
            headers=headers,
        )
        retval = self._content_to_obj(response, self.user_invitation_class)
        return retval, response

    def handle_retrieve_invite_user_constraints(
        self,
        cmd: command.RetrieveInviteUserConstraintsCommand,
        route_prefix: str,
        headers: dict[str, str] | None,
    ) -> tuple[Any, Response]:
        response = self.test_client.get(
            route_prefix + "/invite_user/constraints",
            headers=headers,
        )
        retval = self._content_to_obj(response, self.user_invitation_constraints_class)
        return retval, response

    def handle_register_invited_user(
        self,
        cmd: command.RegisterInvitedUserCommand,
        route_prefix: str,
        headers: dict[str, str] | None,
    ) -> tuple[Any, Response]:
        response = self.test_client.post(
            route_prefix + f"/user_registrations/{cmd.token}",
            headers=headers,
        )
        retval = self._content_to_obj(response, self.user_class)
        return retval, response

    def handle_update_user(
        self,
        cmd: command.UpdateUserCommand,
        route_prefix: str,
        headers: dict[str, str] | None,
    ) -> tuple[Any, Response]:
        request_body = self.update_user_request_body(
            is_active=cmd.is_active,
            roles=cmd.roles,
            organization_id=cmd.organization_id,
        )
        cmd_dict = json.loads(cmd.model_dump_json())
        tgt_user_id = cmd_dict["tgt_user_id"]
        response = self.test_client.put(
            route_prefix + f"/users/{tgt_user_id}",
            headers=headers,
            json=json.loads(request_body.model_dump_json()),
        )
        retval = self._content_to_obj(response, self.user_class)
        return retval, response

    def get_headers(self, cmd: Command, **kwargs: Any) -> dict[str, str] | None:
        if cmd.user:
            assert cmd.user is not None
            headers = self.get_dummy_jwt_header(cmd.user.get_key())
        else:
            headers = None
        return headers

    def get_dummy_jwt(
        self,
        email: str,
        iss: str | None = None,
        sub: str | None = None,
        aud: str | None = None,
        exp: int | None = None,
        expire_default_minutes: int = 15,
    ) -> str:
        claims = {
            "__key__": email,
            "email": email,
            "iss": iss or f"https://{uuid.uuid4()}.org",
            "sub": sub or str(uuid.uuid4()),
            "aud": aud or str(uuid.uuid4()),
            "exp": exp or datetime.now() + timedelta(minutes=expire_default_minutes),
        }
        encoded_jwt = jwt.encode(
            claims, self.SECRET_KEY, algorithm=self.ENCRYPTION_ALGORITHM
        )
        return encoded_jwt

    def get_dummy_jwt_header(
        self,
        email: str,
        iss: str | None = None,
        sub: str | None = None,
        aud: str | None = None,
        exp: int | None = None,
        expire_default_minutes: int = 15,
    ) -> dict[str, str]:
        return {
            "Authorization": f"Bearer {self.get_dummy_jwt(email, iss, sub, aud, exp, expire_default_minutes)}"
        }

    def handle_crud_command(
        self, cmd: CrudCommand, route_prefix: str, headers: dict[str, str] | None
    ) -> tuple[Any, Response]:
        model_class = cmd.MODEL_CLASS
        entity = model_class.ENTITY
        assert entity is not None
        route = f"{route_prefix}/{entity.snake_case_plural_name}"
        if cmd.operation == CrudOperation.READ_ALL:
            if cmd.query_filter:
                response = self.test_client.post(
                    route + "/query",
                    json=json.loads(cmd.query_filter.model_dump_json()),
                    headers=headers,
                )
            else:
                response = self.test_client.get(route, headers=headers)
            retval = self._content_to_obj(response, model_class, is_list=True)
        elif cmd.operation == CrudOperation.READ_SOME:
            assert isinstance(cmd.obj_ids, list)
            ids = json.dumps([str(x) for x in cmd.obj_ids])
            response = self.test_client.get(
                f"{route}/batch",
                headers=headers,
                params={"ids": ids},
            )
            retval = self._content_to_obj(response, model_class, is_list=True)
        elif cmd.operation == CrudOperation.READ_ONE:
            response = self.test_client.get(
                f"{route}/{cmd.obj_ids}",
                headers=headers,
            )
            retval = self._content_to_obj(response, model_class)
        elif cmd.operation == CrudOperation.CREATE_ONE:
            assert isinstance(cmd.objs, model.Model)
            response = self.test_client.post(
                f"{route}",
                json=json.loads(cmd.objs.model_dump_json()),
                headers=headers,
            )
            retval = self._content_to_obj(response, model_class)
        elif cmd.operation == CrudOperation.CREATE_SOME:
            assert isinstance(cmd.objs, list)
            response = self.test_client.post(
                f"{route}/batch",
                json=[json.loads(x.model_dump_json()) for x in cmd.objs],
                headers=headers,
            )
            retval = self._content_to_obj(response, model_class, is_list=True)
        elif cmd.operation == CrudOperation.UPDATE_ONE:
            assert isinstance(cmd.objs, model.Model)
            response = self.test_client.put(
                f"{route}/{cmd.objs.id}",
                json=json.loads(cmd.objs.model_dump_json()),
                headers=headers,
            )
            retval = self._content_to_obj(response, model_class)
        elif cmd.operation == CrudOperation.UPDATE_SOME:
            assert isinstance(cmd.objs, list)
            response = self.test_client.put(
                f"{route}",
                json=[json.loads(x.model_dump_json()) for x in cmd.objs],
                headers=headers,
            )
            retval = self._content_to_obj(response, model_class, is_list=True)
        elif cmd.operation == CrudOperation.DELETE_ONE:
            assert isinstance(cmd.obj_ids, UUID)
            response = self.test_client.delete(
                f"{route}/{cmd.obj_ids}", headers=headers
            )
            retval = self._content_to_obj(response, UUID)
        elif cmd.operation == CrudOperation.DELETE_SOME:
            assert isinstance(cmd.obj_ids, list)
            ids = json.dumps([str(x) for x in cmd.obj_ids])
            response = self.test_client.delete(
                f"{route}/batch",
                headers=headers,
                params={"ids": ids},
            )
            retval = self._content_to_obj(response, UUID, is_list=True)
        else:
            raise NotImplementedError(f"Unsupported operation: {cmd.operation}")
        return retval, response

    @staticmethod
    def _content_to_obj(
        response: Response, retval_class: type, is_list: bool = False
    ) -> Any:
        if response.status_code not in (200, 201):
            return None
        decoded_obj = json.loads(response.content.decode(response.encoding or "utf-8"))
        if issubclass(retval_class, PydanticBaseModel):
            if is_list:
                return [retval_class(**x) for x in decoded_obj]
            else:
                return retval_class(**decoded_obj)
        elif issubclass(retval_class, UUID):
            if is_list:
                return [UUID(x) for x in decoded_obj]
            else:
                return UUID(decoded_obj)
        raise NotImplementedError(f"Unsupported return type: {retval_class}")



import traceback

# pylint: disable=unused-import-alias
from collections.abc import Callable, Iterable
from enum import Enum
from typing import Any

from dynaconf import Dynaconf

from gen_epix import fastapp
from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.commondb.base_env import BaseAppComposer
from gen_epix.commondb.config import AppCfg
from gen_epix.commondb.domain import DOMAIN, command, enum, model
from gen_epix.commondb.domain.model import SORTED_SERVICE_TYPES
from gen_epix.commondb.domain.policy.permission import RoleGenerator
from gen_epix.commondb.services import AuthService, RbacService
from gen_epix.commondb.services.abac import AbacService
from gen_epix.commondb.services.organization import OrganizationService
from gen_epix.commondb.services.system import SystemService
from gen_epix.commondb.services.user_manager import UserManager
from gen_epix.fastapp.domain.domain import Domain
from gen_epix.fastapp.repository import BaseRepository
from gen_epix.fastapp.service import BaseService
from gen_epix.fastapp.util import create_ssl_context


class App(fastapp.App):
    """
    Application class for the GenEpix FastAPI application. Overrides some properties
    to provide more specific types and linter support.
    """

    @property
    def cfg(self) -> Dynaconf:
        return super().cfg


class AppComposer(BaseAppComposer):
    def __init__(
        self,
        app_cfg: AppCfg,
        domain: Domain | None = None,
        sorted_service_types: tuple[Enum] | None = None,
        role_generator_class: type[RoleGenerator] | None = None,
        rbac_service_class: type[RbacService] | None = None,
        user_manager_class: type[UserManager] | None = None,
        model_class_map: dict[type[model.Model], type[model.Model]] | None = None,
        command_class_map: (
            dict[type[command.Command], type[command.Command]] | None
        ) = None,
        policy_class_map: (
            dict[type[fastapp.Policy], type[fastapp.Policy]] | None
        ) = None,
        log_setup: bool = True,
        **kwargs: Any,
    ):
        # Parse input
        self._app_cfg = app_cfg
        self._cfg = app_cfg.cfg
        self._domain = domain or DOMAIN
        self._sorted_service_types = sorted_service_types or SORTED_SERVICE_TYPES
        self._role_generator_class = role_generator_class or RoleGenerator
        self._rbac_service_class = rbac_service_class or RbacService
        self._user_manager_class = user_manager_class or UserManager
        self._model_class_map = model_class_map or {}
        self._command_class_map = command_class_map or {}
        self._policy_class_map = policy_class_map or {}
        self._log_setup = log_setup

        # Derive some properties
        self._role_map = self._role_generator_class.get_role_map()
        self._role_set_map = self._role_generator_class.get_role_set_map()
        self._role_permissions_map = (
            self._role_generator_class.get_role_permissions_map()
        )

        # Compose application
        data = self.compose_application(**kwargs)
        self._app: App = data["app"]
        self._services: dict[Enum, BaseService] = data["services"]
        self._repositories: dict[Enum, BaseRepository] = data["repositories"]
        self._registered_user_dependency: Callable = data["registered_user_dependency"]
        self._new_user_dependency: Callable = data["new_user_dependency"]
        self._idp_user_dependency: Callable = data["idp_user_dependency"]

    def compose_application(self) -> dict:

        # Get loggers
        cfg = self._app_cfg.cfg
        setup_logger = self._app_cfg.setup_logger
        app_logger = self._app_cfg.app_logger
        service_logger = self._app_cfg.service_logger

        # Compose application
        try:
            if self._log_setup and setup_logger:
                setup_logger.debug(
                    App.create_static_log_message(
                        "e8665136", "Starting composing application"
                    )
                )

                setup_logger.debug(
                    App.create_static_log_message(
                        "fb612692", "Initialising services and repositories"
                    )
                )

            # Initialize app
            app_impl = AppImplDetails(
                sorted_service_types=list(self._sorted_service_types),
                rbac_service_class=self._rbac_service_class,
                user_manager_class=self._user_manager_class,
                model_class_map=self._model_class_map,
                command_class_map=self._command_class_map,
                policy_class_map=self._policy_class_map,
                role_map=self._role_map,
                role_set_map=self._role_set_map,
                role_permissions_map=self._role_permissions_map,
            )
            app = App(
                name=self._app_cfg.app_name,
                domain=self._domain,
                cfg=self._app_cfg.cfg,
                impl=app_impl,
                logger=app_logger if self._log_setup else None,
                id_factory=cfg["service"]["defaults"]["props"]["id_factory"],
            )
            ssl_context = create_ssl_context(
                host=cfg["app"]["host"], ssl_cert_file=cfg["app"].get("ssl_cert_file")
            )

            # Initialise repositories and services
            for service_type in self._sorted_service_types:
                service_cfg = cfg["service"][service_type.value]
                service_class = service_cfg["class"]
                service_props = service_cfg["props"]
                repository_cfg = cfg["repository"].get(service_type.value)

                # Create repository if necessary
                curr_repository = None
                if repository_cfg:
                    repository_class: type[BaseRepository] = repository_cfg["class"]
                    repository_props = repository_cfg["props"]
                    if isinstance(repository_cfg["type"], str):
                        repository_type = enum.RepositoryType(repository_cfg["type"])
                    else:
                        repository_type = enum.RepositoryType(
                            repository_cfg["type"].value
                        )
                    entities = app.domain.get_dag_sorted_entities(
                        service_type=service_type
                    )
                    if self._log_setup and setup_logger:
                        setup_logger.debug(
                            app.create_log_message(
                                "db89f0a5",
                                f"Setting up {service_type.value} service with {repository_type.value} repository",
                            )
                        )
                    curr_repository = repository_class.create_repository(
                        entities=entities, **repository_props
                    )
                    # Add to overview of repositories
                    app_impl.repositories[service_type] = curr_repository

                # Create service, injecting app, repository, logger and props
                curr_service: BaseService = service_class(
                    app,
                    service_type=service_type,
                    repository=curr_repository,
                    logger=service_logger,
                    setup_logger=setup_logger if self._log_setup else None,
                    name=service_type.value,
                    ssl_context=ssl_context,
                    **service_props,
                )
                # Add to overview of services
                app_impl.services[service_type] = curr_service

            # Get common services and types
            system_service_type = AppComposer._get_enum_from_list(
                self._sorted_service_types, "SYSTEM"
            )
            system_service = app_impl.services[system_service_type]
            assert isinstance(system_service, SystemService)
            auth_service_type = AppComposer._get_enum_from_list(
                self._sorted_service_types, "AUTH"
            )
            auth_service = app_impl.services[auth_service_type]
            assert isinstance(auth_service, AuthService)
            rbac_service_type = AppComposer._get_enum_from_list(
                self._sorted_service_types, "RBAC"
            )
            rbac_service = app_impl.services[rbac_service_type]
            assert isinstance(rbac_service, RbacService)
            abac_service_type = AppComposer._get_enum_from_list(
                self._sorted_service_types, "ABAC"
            )
            abac_service = app_impl.services[abac_service_type]
            assert isinstance(abac_service, AbacService)
            organization_service_type = AppComposer._get_enum_from_list(
                self._sorted_service_types, "ORGANIZATION"
            )
            organization_service = app_impl.services[organization_service_type]
            assert isinstance(organization_service, OrganizationService)

            # Set up roles
            rbac_service.register_roles(
                app_impl.role_permissions_map, app_impl.role_map[enum.Role.ROOT]
            )

            # Create and set user generator, which can create new users under different scenarios
            # such as from claims, from invitation, and when matching root secret
            app.user_manager = app_impl.user_manager_class(
                organization_service,
                rbac_service,
                root_cfg=app.cfg["service"]["auth"]["props"]["root"],
                automatic_new_user_cfg=app.cfg["service"]["auth"]["props"][
                    "automatic_new_user"
                ],  # set to None if no automatic new user
            )

            # Get current user and new user dependencies for injecting authentication in endpoints
            (
                app_impl.registered_user_dependency_or_none,
                app_impl.new_user_dependency_or_none,
                app_impl.idp_user_dependency_or_none,
            ) = auth_service.create_user_dependencies()

            # Register policies with app
            if self._log_setup and setup_logger:
                setup_logger.debug(
                    app.create_log_message("f329be4d", "Registering policies")
                )
            system_service.register_policies()
            rbac_service.register_policies()
            abac_service.register_policies()

            # Finalise process
            if self._log_setup and setup_logger:
                setup_logger.debug(
                    app.create_log_message("da172304", "Finished composing application")
                )

        except Exception as e:

            # Print error for deployment log, in regular log is not shown there
            traceback.print_exc()
            if self._log_setup and setup_logger:
                setup_logger.error(
                    App.create_static_log_message(
                        "db960800",
                        f"Error setting up application: {e}",
                    )
                )
            raise e

        return {
            "app": app,
            "services": app_impl.services,
            "repositories": app_impl.repositories,
            "registered_user_dependency": app_impl.registered_user_dependency,
            "new_user_dependency": app_impl.new_user_dependency,
            "idp_user_dependency": app_impl.idp_user_dependency,
        }

    @staticmethod
    def _get_enum_from_list(enums: Iterable[Enum], name: str) -> Enum:
        for enum_item in enums:
            if enum_item.name == name:
                return enum_item
        raise ValueError(f"Enum with name {name} not found")



from typing import Any

from sqlalchemy import Engine, select

from gen_epix.commondb.domain import model
from gen_epix.commondb.domain.repository.organization import BaseOrganizationRepository
from gen_epix.commondb.repositories.sa_model.organization import User, UserInvitation
from gen_epix.fastapp import BaseUnitOfWork, CrudOperation, exc
from gen_epix.fastapp.repositories import SARepository
from gen_epix.fastapp.repositories.sa.unit_of_work import SAUnitOfWork


class OrganizationSARepository(SARepository, BaseOrganizationRepository):
    def __init__(
        self,
        engine: Engine,
        user_class: type[model.User] = model.User,
        user_invitation_class: type[model.UserInvitation] = model.UserInvitation,
        sa_user_class: type[User] = User,
        sa_user_invitation_class: type[UserInvitation] = UserInvitation,
        **kwargs: Any,
    ):
        self.sa_user_class = sa_user_class
        self.sa_user_invitation_class = sa_user_invitation_class
        BaseOrganizationRepository.__init__(
            self, user_class=user_class, user_invitation_class=user_invitation_class
        )
        SARepository.__init__(self, engine, **kwargs)

    def is_existing_user_by_key(
        self, uow: BaseUnitOfWork, user_key: str | None
    ) -> bool:
        if user_key is None:
            return False
        assert isinstance(uow, SAUnitOfWork)
        user_row = uow.session.execute(
            select(self.sa_user_class.id).where(self.sa_user_class.key == user_key.lower())
        ).all()
        return True if user_row else False

    def retrieve_user_by_key(self, uow: BaseUnitOfWork, user_key: str) -> model.User:
        # TODO: add filter to crud method instead of retrieving all users
        users: list[model.User] = self.crud(  # type:ignore[assignment]
            uow,
            None,
            self.user_class,
            None,
            None,
            CrudOperation.READ_ALL,
        )
        for user in users:
            if user.key == user_key.lower():
                return user
        raise exc.NoResultsError()



from enum import Enum


class TestType(Enum):
    UNIT = "UNIT"
    INTEGRATION = "INTEGRATION"
    PERFORMANCE = "PERFORMANCE"
    OTHER = "OTHER"
    UNDEFINED = "UNDEFINED"


class RepositoryType(Enum):
    DICT = "DICT"
    SA_SQLITE = "SA_SQLITE"
    SA_SQL = "SA_SQL"



from collections.abc import Callable, Iterable
from typing import Any, NoReturn
from uuid import UUID

from fastapi import APIRouter, FastAPI
from fastapi.responses import StreamingResponse
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from gen_epix.commondb.app_impl_details import AppImplDetails
from gen_epix.fastapp import App
from gen_epix.fastapp.api import CrudEndpointGenerator
from gen_epix.seqdb.domain import command, enum, model


class RetrievePhylogeneticTreeRequestBody(PydanticBaseModel):
    seq_distance_protocol_id: UUID
    tree_algorithm: enum.TreeAlgorithm
    seq_ids: list[UUID]
    leaf_codes: list[str] | None = None


class RetrieveSeqRequestBody(PydanticBaseModel):
    seq_ids: list[UUID]


class RetrieveSeqFastaRequestBody(PydanticBaseModel):

    seq_ids: list[UUID] = Field(
        description="List of sequence IDs to retrieve in FASTA format.",
    )

    file_name: str = Field(
        description="The desired filename for the FASTA download.",
    )


class RetrieveAlleleProfileRequestBody(PydanticBaseModel):
    seq_ids: list[UUID]
    locus_set_id: UUID


def create_seq_endpoints(
    router: APIRouter | FastAPI,
    app: App,
    handle_exception: Callable[[str, Any, Exception], NoReturn] | None = None,
    **kwargs: Any,
) -> None:
    assert handle_exception
    app_impl: AppImplDetails = app.impl
    registered_user_dependency = app_impl.registered_user_dependency

    @router.post(
        "/retrieve/phylogenetic_tree",
        operation_id="retrieve__phylogenetic_tree",
        name="RetrievePhylogeneticTree",
        description=command.RetrievePhylogeneticTreeCommand.__doc__,
    )
    async def retrieve__phylogenetic_tree(
        # user: registered_user_dependency, request_body: RetrievePhylogeneticTreeRequestBody  # type: ignore
        user: registered_user_dependency,
        request_body: RetrievePhylogeneticTreeRequestBody,  # type: ignore
    ) -> model.PhylogeneticTree:
        try:
            retval: model.PhylogeneticTree = app.handle(
                command.RetrievePhylogeneticTreeCommand(
                    user=user,
                    seq_distance_protocol_id=request_body.seq_distance_protocol_id,
                    tree_algorithm=request_body.tree_algorithm,
                    seq_ids=request_body.seq_ids,
                    leaf_names=request_body.leaf_codes,
                )
            )
        except Exception as exception:
            handle_exception("dc71bce0", user, exception, request_ids=request_body.seq_ids)  # type: ignore
        return retval

    @router.post(
        "/retrieve/seq",
        operation_id="retrieve__seq",
        name="RetrieveSeq",
        description=command.RetrieveCompleteSeqCommand.__doc__,
    )
    async def retrieve__seq(
        user: registered_user_dependency, request_body: RetrieveSeqRequestBody  # type: ignore
    ) -> list[model.CompleteSeq]:
        try:
            retval: list[model.CompleteSeq] = app.handle(
                command.RetrieveCompleteSeqCommand(
                    user=user,
                    seq_ids=request_body.seq_ids,
                )
            )
        except Exception as exception:
            handle_exception("ac218f73", user, exception, request_ids=request_body.seq_ids)  # type: ignore
        return retval

    @router.post(
        "/retrieve/seq_fasta",
        operation_id="retrieve__seq_fasta",
        name="RetrieveSeqFasta",
        description=command.RetrieveSeqFastaCommand.__doc__,
    )
    async def retrieve__seq_fasta(
        user: registered_user_dependency, request_body: RetrieveSeqFastaRequestBody
    ) -> StreamingResponse:
        try:
            fasta_iterable: Iterable[str] = app.handle(
                command.RetrieveSeqFastaCommand(
                    user=user,
                    seq_ids=request_body.seq_ids,
                )
            )
        except Exception as exception:
            handle_exception(
                "e4f3b8c1", user, exception, request_ids=request_body.seq_ids
            )

        return StreamingResponse(
            fasta_iterable,
            media_type="application/x-fasta",
            headers={
                "Content-Disposition": f'attachment; filename="{request_body.file_name}"'
            },
        )

    @router.post(
        "/retrieve/allele_profile",
        operation_id="retrieve__allele_profile",
        name="RetrieveAlleleProfile",
        description=command.RetrieveCompleteAlleleProfileCommand.__doc__,
    )
    async def retrieve__allele_profile(
        user: registered_user_dependency, request_body: RetrieveAlleleProfileRequestBody  # type: ignore
    ) -> list[model.CompleteAlleleProfile]:
        try:
            retval: list[model.CompleteAlleleProfile] = app.handle(
                command.RetrieveCompleteAlleleProfileCommand(
                    user=user,
                    seq_ids=request_body.seq_ids,
                    locus_set_id=request_body.locus_set_id,
                )
            )
        except Exception as exception:
            handle_exception("f1d282b4", user, exception, request_ids=request_body.seq_ids)  # type: ignore
        return retval

    # CRUD
    crud_endpoint_sets = CrudEndpointGenerator.create_crud_endpoint_set_for_domain(
        app,
        service_type=enum.ServiceType.SEQ,
        user_dependency=registered_user_dependency,
    )
    CrudEndpointGenerator.generate_endpoints(
        router, crud_endpoint_sets, handle_exception
    )



import base64
import json
import logging
import ssl
import time
import urllib.parse
from typing import Any, TypedDict
from uuid import UUID

import httpx
import jwt
from fastapi import Request
from fastapi.openapi.models import OAuthFlowAuthorizationCode, OAuthFlows, SecurityBase
from fastapi.security import OAuth2

# from fastapi.openapi.models import OAuth2, OAuthFlowAuthorizationCode, OAuthFlows
from fastapi.security.open_id_connect_url import OpenIdConnect
from fastapi.security.utils import get_authorization_scheme_param

from gen_epix.fastapp import exc
from gen_epix.fastapp.enum import AuthProtocol, OAuthFlow
from gen_epix.fastapp.log import BaseLogItem, LogItem
from gen_epix.fastapp.services.auth.idp_client import IdpClient
from gen_epix.fastapp.services.auth.model import Claims, IdentityProvider, OidcServerCfg


class OauthIdpClient(IdpClient, OpenIdConnect):

    DEFAULT_INTROSPECTION_REQUEST_HEADERS: dict[str, str] = {
        "Content-Type": "application/x-www-form-urlencoded",
    }
    DEFAULT_INTROSPECTION_AUTH_METHOD: str = "client_secret_basic"
    DEFAULT_CLIENT_CREDENTIAL_FLOW_REQUEST_HEADERS: dict[str, str] = {
        "Content-Type": "application/x-www-form-urlencoded",
    }
    DEFAULT_CLIENT_CREDENTIAL_FLOW_MAX_RETRIES: int = 3
    DEFAULT_CLIENT_CREDENTIAL_FLOW_BASE_DELAY: float = 1.0  # in seconds
    DEFAULT_ALLOWED_SIGNING_ALGORITHMS: list[str] = ["RS256"]

    def __init__(
        self,
        server_cfg: OidcServerCfg,
        token_name: str | None = None,
        logger: logging.Logger | None = None,
        log_item_class: type[BaseLogItem] = LogItem,
        discovery_url: str | None = None,
        discovery_doc: dict[str, Any] | None = None,
        id: UUID | None = None,
        ssl_context: ssl.SSLContext | bool = True,
        introspect_token_request_headers: dict[str, str] | None = None,
        client_credential_flow_request_headers: dict[str, str] | None = None,
        client_credential_flow_max_retries: int | None = None,
        client_credential_flow_base_delay: float | None = None,
        **kwargs: Any,
    ):
        # Set IdpClient properties
        issuer = server_cfg.issuer
        if issuer is None:
            # Fetch issuer later from discovery document
            issuer = ""
        super().__init__(
            issuer,
            token_name=token_name or self.DEFAULT_TOKEN,
            id=id,
            ssl_context=ssl_context,
            **kwargs,
        )

        # Set input properties
        self.server_cfg = server_cfg.model_copy()
        self.logger = logger
        self._log_item_class = log_item_class
        self._signing_keys: dict[str, jwt.PyJWK] = {}
        self._introspection_request_headers = (
            introspect_token_request_headers
            or self.DEFAULT_INTROSPECTION_REQUEST_HEADERS
        )
        self._client_credential_flow_request_headers = (
            client_credential_flow_request_headers
            or self.DEFAULT_CLIENT_CREDENTIAL_FLOW_REQUEST_HEADERS
        )
        self._client_credential_flow_max_retries = (
            client_credential_flow_max_retries
            or self.DEFAULT_CLIENT_CREDENTIAL_FLOW_MAX_RETRIES
        )
        self._client_credential_flow_base_delay = (
            client_credential_flow_base_delay
            or self.DEFAULT_CLIENT_CREDENTIAL_FLOW_BASE_DELAY
        )
        self._allowed_signing_algorithms = (
            self.server_cfg.id_token_signing_alg_values_supported
            or self.DEFAULT_ALLOWED_SIGNING_ALGORITHMS
        )

        # Introspection cache entry type
        class IntrospectionCacheEntry(TypedDict):
            active: bool | None
            last_checked: int
            exp: int

        self._introspection_cache: dict[str, IntrospectionCacheEntry] = {}

        # TODO: @cached(cache=TTLCache(maxsize=1024, ttl=300)) ? ?
        if self.server_cfg.enable_introspection:
            # dynamically determine the introspection endpoint
            self.introspection_endpoint: str = self._get_introspection_endpoint()

        # Set cfg and retrieve remaining information
        self.update_server_config_from_discovery(url=discovery_url, doc=discovery_doc)
        if issuer == "":
            self.scheme_name = self.server_cfg.issuer or ""

        # Set SecurityBase properties
        authorization_endpoint = (
            self.server_cfg.authorization_endpoint or ""
        )  # In case of client credentials flow or development, this may not be set
        token_endpoint = (
            self.server_cfg.token_endpoint or ""
        )  # In case of client credentials flow or development, this may not be set
        flows = OAuthFlows()
        flows.authorizationCode = OAuthFlowAuthorizationCode(
            authorizationUrl=authorization_endpoint,
            tokenUrl=token_endpoint,
            scopes={x: x for x in self.server_cfg.scope.split()},
        )
        self.model: SecurityBase = OAuth2(flows=flows)

    @property
    def issuer(self) -> str:
        assert self.server_cfg.issuer is not None
        return self.server_cfg.issuer

    @property
    def audience(self) -> str:
        return self.server_cfg.audience or self.server_cfg.client_id

    @property
    def scope(self) -> str:
        assert self.server_cfg.scope is not None
        return self.server_cfg.scope

    def update_server_config_from_discovery(
        self,
        url: str | None = None,
        doc: dict[str, Any] | None = None,
    ) -> None:
        """
        Update the OIDC configuration from the discovery URL or, if provided, the
        discovery document.
        """
        url = url or self.server_cfg.discovery_url
        if url is None and doc is None:
            raise exc.InitializationServiceError(
                "No discovery URL or document provided for OIDC configuration"
            )

        # Special case: discovery document provided -> update from that first
        if doc:
            # Update current configuration from provided discovery document
            for key, value in doc.items():
                setattr(self.server_cfg, key, value)

        # Update from discovery URL
        if not url:
            return
        try:
            # Get discovery document
            with httpx.Client(verify=self.ssl_context) as client:
                response = client.get(url)
                discovery_doc = response.json()

            # Update current configuration with discovery data, preserving client credentials
            for key, value in discovery_doc.items():
                if (
                    key not in OidcServerCfg.NON_SPEC_FIELDS
                    and key in self.server_cfg.__class__.model_fields
                ):
                    setattr(self.server_cfg, key, value)

            if not self.server_cfg.is_valid():
                invalid_fields = self.server_cfg.get_invalid_fields()
                raise exc.InitializationServiceError(
                    f"OIDC configuration from discovery URL is not valid. Invalid fields: {invalid_fields}"
                )
        except Exception as exception:
            if self.logger:
                self.logger.error(
                    self._log_item_class(
                        code="cfe970aa",
                        msg="Error accessing discovery URL",
                        scheme_name=self.server_cfg.name,
                        exception=exception,
                    ).dumps()
                )
            raise exc.InitializationServiceError(msg) from exception

    async def get_jwk_from_jwt(self, jwt_token: str) -> jwt.PyJWK:
        try:
            header = jwt.get_unverified_header(jwt_token)
        except jwt.PyJWTError as e:
            if self.logger:
                self.logger.warning(
                    self._log_item_class(
                        code="4cff1367",
                        msg="Unable to parse header from token",
                        scheme_name=self.scheme_name,
                        jwt=jwt_token,
                        exception=e,
                    ).dumps()
                )
            raise exc.UnauthorizedAuthError() from e

        key_id = header.get("kid")
        if not key_id:
            if self.logger:
                self.logger.warning(
                    self._log_item_class(
                        code="0184bc35",
                        msg="No key ID found in token header",
                        scheme_name=self.scheme_name,
                        jwt=jwt_token,
                    ).dumps()
                )
            raise exc.UnauthorizedAuthError()

        # Verify that the signing key in this session is outdated, fetch new one if so
        # TODO: verify if fetching new signing keys is ok
        key = self._signing_keys.get(key_id)
        if not key:
            if self.logger and self.logger.level <= logging.DEBUG:
                self.logger.debug(
                    self._log_item_class(
                        code="e90dd1aa",
                        msg="Key ID not found among signing keys, fetching new ones",
                        scheme_name=self.scheme_name,
                    ).dumps()
                )
            self._load_keys()
            key = self._signing_keys.get(key_id)
            if not key:
                if self.logger:
                    self.logger.warning(
                        self._log_item_class(
                            code="2a5975ff",
                            msg="Key ID not found amoung newly fetched signing keys",
                            scheme_name=self.scheme_name,
                            key_id=key_id,
                        ).dumps()
                    )
                raise exc.UnauthorizedAuthError()
            if self.logger and self.logger.level <= logging.DEBUG:
                self.logger.debug(
                    self._log_item_class(
                        code="c448ead5",
                        msg="Key ID found among newly fetched signing keys",
                        scheme_name=self.scheme_name,
                    ).dumps()
                )
        return key

    async def get_claims_from_jwt(
        self, jwt_token: str
    ) -> dict[str, str | int | bool | list[str]] | None:
        # Decode token without verifying signature to make sure this token is generated
        # by this OIDC server
        claims = jwt.decode(jwt_token, options={"verify_signature": False})
        server_cfg = self.server_cfg

        if claims["iss"] != server_cfg.issuer:
            if self.logger and self.logger.level <= logging.DEBUG:
                self.logger.debug(
                    self._log_item_class(
                        code="7e2a1c4d",
                        msg="JWT issuer does not match OIDC server configuration",
                        scheme_name=self.scheme_name,
                        token_issuer=claims["iss"],
                        token_subject=claims.get("sub"),
                        expected_issuer=server_cfg.issuer,
                    ).dumps()
                )
            return None

        # Get key to verify signature and decode again
        key = await self.get_jwk_from_jwt(jwt_token)
        try:
            claims = jwt.decode(
                jwt_token,
                key=key,
                algorithms=self._allowed_signing_algorithms,
                audience=self.audience,
                issuer=server_cfg.issuer,
                options={
                    "require_iat": True,
                    "verify_iat": True,
                    "require_exp": True,
                    "verify_exp": True,
                },
            )
        except Exception as exception:
            msg = "Unable to decode JWT: "

            if isinstance(exception, jwt.ExpiredSignatureError):
                msg += "signature has expired"
            elif isinstance(exception, jwt.PyJWTError):
                msg += "signature is invalid"
            else:
                msg += "unknown issue"
            if self.logger:
                self.logger.warning(
                    self._log_item_class(
                        code="f4b73564",
                        msg=msg,
                        scheme_name=self.scheme_name,
                        exception=exception,
                    ).dumps()
                )
            raise exc.CredentialsAuthError(
                http_props={"headers": {"WWW-Authenticate": "Bearer"}}
            ) from exception

        # optionally apply token introspection
        if self.server_cfg.enable_introspection:
            self.introspect_token(jwt_token, claims)

        # Get issuer and sub
        issuer = claims["iss"]
        sub = claims.get("sub")
        if not issuer or not sub:
            if not issuer and not sub:
                msg_part = "no issuer and no sub"
            elif issuer and not sub:
                msg_part = "no sub"
            else:
                msg_part = "no issuer"
            if self.logger:
                self.logger.warning(
                    self._log_item_class(
                        code="b4a1d49b",
                        msg=f"JWT does not contain required claims: {msg_part}",
                        scheme_name=self.scheme_name,
                    ).dumps()
                )
            raise exc.CredentialsAuthError(
                http_props={"headers": {"WWW-Authenticate": "Bearer"}}
            )

        # Log for debugging
        if self.logger and self.logger.level <= logging.DEBUG:
            self.logger.debug(
                self._log_item_class(
                    code="8a7c4e92",
                    msg="JWT is valid",
                    scheme_name=self.scheme_name,
                    token_issuer=claims["iss"],
                ).dumps()
            )

        # Map claims according to claim_map, allowing e.g. to standardize claim names across IDPs
        for new_claim_name, orig_claim_names in server_cfg.claim_map.items():
            for orig_claim_name in orig_claim_names:
                claims[new_claim_name] = claims.get(orig_claim_name)
                if claims[new_claim_name] is not None:
                    break

        return claims

    def introspect_token(self, jwt_token: str, claims: dict[str, Any]) -> None:
        now = self._now()
        self._prune_expired_introspection_cache(now)
        if self._is_cached_introspection_token_inactive(jwt_token):
            if self.logger:
                self.logger.warning(
                    self._log_item_class(
                        code="0ce44f1a",
                        msg="Token previously marked inactive by introspection; denying",
                    ).dumps()
                )
            raise exc.CredentialsAuthError(
                http_props={"headers": {"WWW-Authenticate": "Bearer"}}
            )
        if self._is_recheck_introspection(jwt_token, now):
            if self.logger:
                self.logger.info(
                    self._log_item_class(
                        code="9deaa6b2",
                        msg="Performing token introspection re-check",
                    ).dumps()
                )
            is_active = self._introspect_token_with_server(jwt_token)
            if is_active is None:
                # network or parse error
                raise exc.CredentialsAuthError(
                    http_props={"headers": {"WWW-Authenticate": "Bearer"}}
                )
            exp_val = int(claims.get("exp", now))
            if is_active:
                self._update_introspection_cache(jwt_token, True, exp_val, now)
            else:
                if is_active is False:
                    self._update_introspection_cache(jwt_token, False, exp_val, now)
                    if self.logger:
                        self.logger.warning(
                            self._log_item_class(
                                code="b1c2d3e4",
                                msg="Token marked inactive by introspection; denying",
                            ).dumps()
                        )
                raise exc.CredentialsAuthError(
                    http_props={"headers": {"WWW-Authenticate": "Bearer"}}
                )

    def retrieve_jwt_with_client_credentials_flow(
        self,
        scope: str,
        headers: dict[str, str] | None = None,
        max_retries: int | None = None,
        base_delay: float | None = None,
    ) -> str:
        """
        Call server to get token through OAuth Client Credentials flow.
        """
        # Parse input
        headers = dict(headers or self._client_credential_flow_request_headers)
        max_retries = max_retries or self._client_credential_flow_max_retries
        base_delay = base_delay or self._client_credential_flow_base_delay

        # Add basic auth header
        headers["Authorization"] = (
            "Basic "
            + base64.b64encode(
                f"{self.server_cfg.client_id}:{self.server_cfg.client_secret}".encode()
            ).decode()
        )

        # Get token endpoint URL
        url = self.server_cfg.token_endpoint
        if not isinstance(url, str):
            # Try to get from discovery document
            if self.logger and self.logger.level <= logging.DEBUG:
                self.logger.debug(
                    self._log_item_class(
                        code="8f3a2b1c",
                        msg=f"Token endpoint URL is not set in OIDC server configuration for server {self.server_cfg.name}, trying to update from discovery URL",
                        scheme_name=self.scheme_name,
                    ).dumps()
                )
            self.update_server_config_from_discovery()
            url = self.server_cfg.token_endpoint
        if not isinstance(url, str):
            raise exc.ServiceUnavailableError("Token endpoint URL is not set")

        # Create request body
        token_data: str = "&".join(
            (
                f"grant_type=client_credentials",
                f"scope={urllib.parse.quote(scope)}",
            )
        )

        # Call server with retries
        last_exception: Exception | None = None
        for attempt in range(max_retries + 1):
            try:
                with httpx.Client(verify=self.ssl_context) as client:
                    response = client.post(
                        url,
                        data=token_data,
                        headers=headers,
                    )
                    response.raise_for_status()
                    token_response = response.json()
                    token: str = token_response["access_token"]
                    return token
            except Exception as exception:
                last_exception = exception
                if self.logger:
                    self.logger.warning(
                        self._log_item_class(
                            code="a7f3e9d2",
                            msg=f"OAuth Client Credentials flow token retrieval attempt {attempt + 1} failed for server {self.server_cfg.name}",
                            scheme_name=self.scheme_name,
                            exception=exception,
                        ).dumps()
                    )
            # Wait before next retry (but not after the last attempt)
            if attempt < max_retries:
                time.sleep(base_delay)

        # All retries failed
        if self.logger:
            self.logger.error(
                self._log_item_class(
                    code="f8a3d7b2",
                    msg=f"OAuth Client Credentials flow token retrieval failed after {max_retries + 1} attempts for server {self.server_cfg.name}",
                    scheme_name=self.scheme_name,
                ).dumps()
            )
        raise exc.ServiceUnavailableError(
            f"Token retrieval failed for server {self.server_cfg.name}: {last_exception}"
        )

    def get_claims_from_userinfo(
        self, access_token: str
    ) -> dict[str, str | int | bool | list[str]]:
        userinfo_endpoint = self.server_cfg.userinfo_endpoint
        assert userinfo_endpoint is not None
        try:
            with httpx.Client(verify=self.ssl_context) as client:
                response = client.get(
                    userinfo_endpoint,
                    headers={"Authorization": f"Bearer {access_token}"},
                )
                claims = json.loads(response.content)
                if not isinstance(claims, dict) or "error" in claims:
                    # Currently e.g. "InvalidAuthenticationToken"
                    if self.logger:
                        self.logger.warning(
                            self._log_item_class(
                                code="ce05d050",
                                msg=f"Unable to get claims from {userinfo_endpoint}: claims contain error",
                                scheme_name=self.scheme_name,
                                claims=claims,
                            ).dumps()
                        )
                    raise exc.ServiceUnavailableError()
                return claims
        except Exception as exception:
            if self.logger:
                self.logger.warning(
                    self._log_item_class(
                        code="ac6c84f7",
                        msg=f"Unable to get claims from {userinfo_endpoint}",
                        scheme_name=self.scheme_name,
                        exception=exception,
                    ).dumps()
                )
            return {}

    def get_identity_provider(self) -> IdentityProvider:
        issuer = self.server_cfg.issuer
        assert issuer is not None
        return IdentityProvider(
            name=self.server_cfg.name,
            label=self.server_cfg.label,
            client_id=self.server_cfg.client_id,
            client_secret=self.server_cfg.client_secret,
            discovery_url=self.server_cfg.discovery_url,
            issuer=issuer,
            auth_protocol=AuthProtocol.OIDC,
            oauth_flow=OAuthFlow.AUTHORIZATION_CODE,
            scope=self.server_cfg.scope,
            public=self.server_cfg.public,
        )

    def _get_introspection_endpoint(self) -> str:
        if self.server_cfg.introspection_endpoint:
            return self.server_cfg.introspection_endpoint
        # dynamically determine from discovery document
        try:
            url: str | None = self.server_cfg.discovery_url
            if not url:
                raise exc.AuthException(
                    "Token introspection was enabled but endpoint URL is not set and could also not be determined from discovery URL"
                )
            with httpx.Client(verify=self.ssl_context) as client:
                response = client.get(url)
                discovery_doc = response.json()
            introspection_endpoint: str = discovery_doc.get("introspection_endpoint")
            return introspection_endpoint
        except Exception as exception:
            if self.logger:
                self.logger.error(
                    self._log_item_class(
                        code="d1234abc",
                        msg="Error accessing discovery URL to determine introspection endpoint",
                        scheme_name=self.scheme_name,
                        exception=exception,
                    ).dumps()
                )
            raise exc.UnauthorizedAuthError(
                http_props={"headers": {"WWW-Authenticate": "Bearer"}}
            ) from exception

    def _introspect_token_with_server(self, jwt_token: str) -> bool | None:
        if not self.introspection_endpoint:
            return None
        timeout_seconds = self.server_cfg.introspection_timeout_seconds
        # Prepare headers and body
        headers = self._introspection_request_headers
        data: str = "&".join(
            (
                f"token={urllib.parse.quote(jwt_token)}",
                f"token_type_hint=access_token",
            )
        )
        method = (
            self.server_cfg.introspection_auth_method
            or self.DEFAULT_INTROSPECTION_AUTH_METHOD
        ).lower()
        if method not in {"client_secret_basic", "client_secret_post", "none"}:
            if self.logger:
                self.logger.warning(
                    self._log_item_class(
                        code="6f4a8e22",
                        msg=(
                            "Unknown introspection auth method; defaulting to client_secret_basic"
                        ),
                        method=method,
                    ).dumps()
                )
            method = "client_secret_basic"
        if method == "client_secret_basic":
            # Add Basic auth
            headers["Authorization"] = (
                "Basic "
                + base64.b64encode(
                    f"{self.server_cfg.client_id}:{self.server_cfg.client_secret}".encode()
                ).decode()
            )
        try:
            # TODO: Fix introspection unauthorized error (currently introspection disabled by default)
            with httpx.Client(
                verify=self.ssl_context, timeout=timeout_seconds
            ) as client:
                response = client.post(
                    self.introspection_endpoint, data=data, headers=headers
                )
            if response.status_code != 200:
                if self.logger:
                    self.logger.warning(
                        self._log_item_class(
                            code="7a3e6d1f",
                            msg=("Token introspection returned non-200 status"),
                            status=response.status_code,
                        ).dumps()
                    )
                raise exc.UnauthorizedAuthError(
                    http_props={"headers": {"WWW-Authenticate": "Bearer"}}
                )
            payload = response.json()
            active = payload.get("active")
            if isinstance(active, bool):
                return active
            # Unexpected payload shape
            return None
        except Exception as exc_:  # broad: treat any failure as unknown
            if self.logger:
                self.logger.warning(
                    self._log_item_class(
                        code="f2b3c9aa",
                        msg=("Token introspection failed (timeout/network/parse)"),
                        exception=exc_,
                    ).dumps()
                )
            raise exc.UnauthorizedAuthError(
                http_props={"headers": {"WWW-Authenticate": "Bearer"}}
            ) from exc_

    def _now(self) -> int:
        return int(time.time())

    def _prune_expired_introspection_cache(self, now: int | None = None) -> None:
        time_stamp = now or self._now()
        expired_keys = [
            x for x, y in self._introspection_cache.items() if y["exp"] <= time_stamp
        ]
        for x in expired_keys:
            self._introspection_cache.pop(x, None)

    def _is_cached_introspection_token_inactive(self, jwt_token: str) -> bool:
        introspection_token = self._introspection_cache.get(jwt_token)
        return bool(introspection_token and introspection_token.get("active") is False)

    def _is_recheck_introspection(self, jwt_token: str, now: int | None = None) -> bool:
        introspection_token = self._introspection_cache.get(jwt_token)
        if not introspection_token:
            return True
        interval = self.server_cfg.introspection_interval_seconds
        last = int(introspection_token.get("last_checked", 0))
        time_stamp = now or self._now()
        return (time_stamp - last) >= interval

    def _update_introspection_cache(
        self,
        jwt_token: str,
        active: bool | None,
        exp: int,
        now: int | None = None,
    ) -> None:
        self._introspection_cache[jwt_token] = {
            "active": active,
            "last_checked": now or self._now(),
            "exp": exp,
        }

    def _load_keys(self) -> None:
        jwks_uri = self.server_cfg.jwks_uri
        assert jwks_uri is not None
        try:
            with httpx.Client(verify=self.ssl_context) as client:
                # get keys
                response = client.get(jwks_uri)
                response.raise_for_status()
                response_dict = response.json()
        except Exception as exception:
            if self.logger:
                self.logger.warning(
                    self._log_item_class(
                        code="edab2e97",
                        msg=f"Unable to load new signing keys from {jwks_uri}",
                        scheme_name=self.scheme_name,
                        exception=exception,
                    ).dumps()
                )
            raise exc.ServiceUnavailableError() from exception

        # verify keys
        self._signing_keys = {}
        for key_data in response_dict["keys"]:
            if (
                key_data.get("use") in ["sig"]
                and key_data.get("kty") == "RSA"
            ):
                self._signing_keys[key_data["kid"]] = jwt.PyJWK.from_dict(key_data)

    async def __call__(self, request: Request) -> Claims | None:  # type: ignore
        """
        Retrieve verified claims for the user based on the request.
        """
        if authorization := request.headers.get("authorization"):
            scheme, token = get_authorization_scheme_param(authorization)
            if scheme.upper() == "BEARER":
                # TODO: check if this is a security risk
                # or whether it should return an error
                try:
                    claims = await self.get_claims_from_jwt(token)
                    if not claims:
                        return None
                    return Claims(
                        claims=claims, scheme=scheme, token=token, idp_client_id=self.id
                    )
                except exc.AuthException as exception:
                    if self.logger:
                        self.logger.warning(
                            self._log_item_class(
                                code="ac521d94",
                                msg="Error retrieving claims from JWT",
                                scheme_name=self.scheme_name,
                                exception=exception,
                            ).dumps()
                        )
                    return None
            else:
                # Authorization scheme not implemented
                if self.logger:
                    self.logger.warning(
                        self._log_item_class(
                            code="ecb88df4",
                            msg=f"Authorization scheme {scheme} not implemented",
                            scheme_name=self.scheme_name,
                        ).dumps()
                    )
                return None
        if self.logger:
            self.logger.warning(
                self._log_item_class(
                    code="e1dad160",
                    msg="No authorisation information provided in header",
                    scheme_name=self.scheme_name,
                ).dumps()
            )
        return None



import sqlalchemy as sa
from sqlalchemy.orm import Mapped

from gen_epix.commondb.repositories.sa_model import get_mixin_mapped_column
from gen_epix.seqdb.domain import model


class CodeMixin:
    code: Mapped[str] = get_mixin_mapped_column(model.CodeMixin, "code", sa.String)


class QualityMixin:
    quality_score: Mapped[float] = get_mixin_mapped_column(
        model.QualityMixin, "quality_score", sa.Float
    )
    quality: Mapped[str] = get_mixin_mapped_column(
        model.QualityMixin, "quality", sa.String
    )


class SeqMixin:
    seq: Mapped[str] = get_mixin_mapped_column(model.SeqMixin, "seq", sa.Text)
    seq_format: Mapped[str] = get_mixin_mapped_column(
        model.SeqMixin, "seq_format", sa.String
    )
    seq_hash_sha256: Mapped[bytes] = get_mixin_mapped_column(
        model.SeqMixin, "seq_hash_sha256", sa.LargeBinary
    )
    length: Mapped[int] = get_mixin_mapped_column(model.SeqMixin, "length", sa.Integer)


class AlignmentMixin:
    aln: Mapped[str] = get_mixin_mapped_column(model.AlignmentMixin, "aln", sa.Text)
    aln_format: Mapped[str] = get_mixin_mapped_column(
        model.AlignmentMixin, "aln_format", sa.String
    )
    aln_hash_sha256: Mapped[bytes] = get_mixin_mapped_column(
        model.AlignmentMixin, "aln_hash_sha256", sa.LargeBinary
    )


class ProtocolMixin:
    code: Mapped[str] = get_mixin_mapped_column(model.ProtocolMixin, "code", sa.String)
    name: Mapped[str] = get_mixin_mapped_column(model.ProtocolMixin, "name", sa.String)
    version: Mapped[str] = get_mixin_mapped_column(
        model.ProtocolMixin, "version", sa.String
    )
    description: Mapped[str] = get_mixin_mapped_column(
        model.ProtocolMixin, "description", sa.Text
    )
    props: Mapped[dict[str, str]] = get_mixin_mapped_column(
        model.ProtocolMixin, "props", sa.JSON
    )



from typing import Any

from pydantic import Field

from gen_epix.filter.base import Filter


class EqualsFilter(Filter):
    value: Any = Field(default=None, description="The value to match.", frozen=True)

    def _match(self, value: Any) -> bool:
        is_match: bool = value == self.value
        return is_match


# No typed version of this filter is needed since the type of the values would be needed as well



from typing import Literal

from pydantic import Field

from gen_epix.filter.enum import FilterType
from gen_epix.filter.equals import EqualsFilter


class EqualsBooleanFilter(EqualsFilter):
    value: bool = Field(description="The boolean value to match.", frozen=True)


class TypedEqualsBooleanFilter(EqualsBooleanFilter):
    type: Literal[FilterType.EQUALS_BOOLEAN.value]



import abc
from collections.abc import Hashable
from enum import Enum
from typing import Any

from gen_epix.fastapp import (
    App,
    BaseService,
    EventTiming,
    Permission,
    PermissionType,
    PermissionTypeSet,
    exc,
)
from gen_epix.fastapp.model import Command, User
from gen_epix.fastapp.services.rbac.policy import RbacPolicy


class BaseRbacService(BaseService):
    """
    Abstract base class for a Role-Based Access Control (RBAC) service. Implement this
    class to provide RBAC functionality to the application. Most methods already have
    an implementation, but the following methods must be implemented by the subclass:
    - retrieve_user_roles: retrieve the roles of a user

    A role is defined as a hashable object such as a string or an Enum, which is
    registered with this service together with the associated Permissions. The service
    can subsequently register a policy for commands to which RBAC applies and which
    checks if the user has the required permission to execute the command through the
    role(s) they have been assigned.
    """

    def __init__(self, app: App, **kwargs: Any):
        super().__init__(app, **kwargs)
        self._permissions_without_rbac: set[Permission] = set()
        self._permissions_by_role: dict[Hashable, set[Permission]] = {}
        self._roles_by_permission: dict[Permission, set[Hashable]] = {
            x: set() for x in self.app.domain.permissions
        }
        self._sub_roles_by_role: dict[Hashable, set[Hashable]] = {}

    @property
    def permissions_without_rbac(self) -> set[Permission]:
        return self._permissions_without_rbac

    @property
    def permissions_by_role(self) -> dict[Hashable, set[Permission]]:
        return self._permissions_by_role

    @property
    def roles_by_permission(self) -> dict[Permission, set[Hashable]]:
        return self._roles_by_permission

    def register_handlers(self) -> None:
        """
        No handlers to register in base implementation. Override as needed.
        """
        pass

    def register_permission_without_rbac(self, permission: Permission) -> None:
        """
        Register a permission that is not subject to RBAC. This can be used for
        permissions or commands that are not associated with a role, but are still used
        in the application.
        """
        roles = self._roles_by_permission[permission]
        if roles:
            roles_str = ", ".join([str(x) for x in roles])
            raise exc.ServiceException(
                f"Permission {permission} has some roles registered: {roles_str}"
            )
        self._permissions_without_rbac.add(permission)

    def unregister_permission_without_rbac(self, permission: Permission) -> None:
        """
        Unregister a permission that is not subject to RBAC. This can be used for
        permissions that are not associated with a role, but are still used in the
        application.
        """
        if permission not in self._permissions_without_rbac:
            raise exc.ServiceException(f"Permission {permission} is not registered")
        self._permissions_without_rbac.remove(permission)

    def register_role(
        self, role: Hashable, permissions: set[Permission], update_role: bool = True
    ) -> None:
        """
        Register a role with the associated permissions, or update an existing role's
        permissions. In case of dynamic roles, i.e. that can be changed at runtime, call
        register_role at runtime in when a role is updated or added. In case of multiple
        objs, a mechanism should be implemented to ensure that the role is
        (eventually) updated or added in all objs.
        """
        # Check if all permissions are registered
        invalid_permissions = permissions - self.app.domain.permissions
        if invalid_permissions:
            invalid_permissions_str = ", ".join([str(x) for x in invalid_permissions])
            raise exc.ServiceException(
                f"Permission(s) {invalid_permissions_str} are not registered"
            )
        # Add to permissions by role
        if role in self._permissions_by_role:
            if not update_role:
                raise exc.ServiceException(f"Role {role} is already registered")
            removed_permissions = self._permissions_by_role[role] - permissions
            for permission in removed_permissions:
                self._roles_by_permission[permission].remove(role)
            permissions = permissions - self._permissions_by_role[role]
        else:
            self._permissions_by_role[role] = permissions
        # Add to roles by permission
        for permission in permissions:
            self._roles_by_permission[permission].add(role)
        # Remove from role hierarchy so that it can be recalculated
        for sub_roles in self._sub_roles_by_role.values():
            if role in sub_roles:
                sub_roles.remove(role)
        self._sub_roles_by_role.pop(role, None)

    def register_roles(
        self,
        role_permissions: dict[
            Hashable, set[Permission | tuple[type[Command], PermissionType]]
        ],
        root_role: Hashable | None = None,
        on_missing_root_permissions: str = "add",
        **kwargs: Any,
    ) -> None:
        """
        Register multiple roles with the associated permissions. This is a convenience
        method to register multiple roles at once. The permissions are given as a set of
        Permission objects or tuples of Command class and PermissionType. The root_role
        is optional and is used to check that the root role has all permissions. The
        keyword arguments are passed to the register_role method.
        """
        on_missing_root_permissions = on_missing_root_permissions.lower()
        if on_missing_root_permissions not in ("add", "raise"):
            raise ValueError(
                f"Invalid value for on_missing_root_permissions: {on_missing_root_permissions}"
            )
        all_permissions: set[Permission] = (
            self.app.domain.permissions  # type:ignore[assignment]
        )
        if root_role and root_role not in role_permissions:
            self.register_role(root_role, all_permissions, **kwargs)
        for role, permissions_or_tuples in role_permissions.items():
            permissions: set[Permission] = {
                (
                    x
                    if isinstance(x, Permission)
                    else self.app.domain.get_permission(x[0], x[1])
                )
                for x in permissions_or_tuples
            }
            if root_role and role == root_role:
                missing_permissions = all_permissions - permissions
                if missing_permissions:
                    if on_missing_root_permissions == "raise":
                        missing_permissions_str = ", ".join(
                            [str(x) for x in missing_permissions]
                        )
                        raise exc.InitializationServiceError(
                            f"Root role {root_role} is missing permissions: {missing_permissions_str}"
                        )
                    elif on_missing_root_permissions == "add":
                        # Add all missing permissions
                        permissions = all_permissions
                    else:
                        raise NotImplementedError(
                            f"Unknown on_missing_root_permissions strategy: {on_missing_root_permissions}"
                        )
            self.register_role(role, permissions, **kwargs)

    def unregister_role(self, role: Hashable) -> None:
        """
        Unregister a role. In case of dynamic roles, i.e. that can be changed at
        runtime, call unregister_role at runtime in when a role is deleted. In case
        of multiple objs, a mechanism should be implemented to ensure that the
        role is (eventually) unregistered in all objs.

        Any RBAC policies for the corresponding permissions will be updated
        automatically.
        """
        if role not in self._permissions_by_role:
            raise exc.ServiceException(f"Role {role} is not registered")
        for permission in self._permissions_by_role[role]:
            self._roles_by_permission[permission].remove(role)
        del self._permissions_by_role[role]

    def verify_roles_exist_for_all_permission(self) -> None:
        """
        Verify that all permissions that are subject to RBAC have at least one role
        registered. This is useful to ensure that no such permission is forgotten to be
        associated with a role.
        """
        missing_permissions = self.app.domain.permissions - {
            x for x in self._roles_by_permission if not self._roles_by_permission[x]
        }
        if missing_permissions:
            missing_permissions_str = ", ".join([str(x) for x in missing_permissions])
            raise exc.ServiceException(
                f"No roles for permission(s) {missing_permissions_str}"
            )

    def get_rbac_permissions_for_command_class(
        self, command_class: type[Command]
    ) -> set[Permission]:
        """
        Get the permissions for a command class that are subject to RBAC, i.e. which are
        registered to be subject to RBAC.
        """
        all_permissions = self.app.domain.get_permissions_for_command(command_class)
        assert isinstance(all_permissions, set)
        return all_permissions - self._permissions_without_rbac

    def get_command_classes_with_rbac(self) -> set[type[Command]]:
        """
        Get all command classes that are subject to RBAC, i.e. for which at least one
        permission is registered to be subject to RBAC.
        """
        permissions = self.app.domain.permissions - self._permissions_without_rbac
        return {self.app.domain.get_command_for_permission(x) for x in permissions}

    def get_sub_roles(self, role: Hashable) -> set[Hashable]:
        """
        Get all sub roles of a role, i.e. all roles whose permissions are a subset of
        the role in question. The role hierarchy is calculated lazily.
        """
        if role not in self._sub_roles_by_role:
            self._sub_roles_by_role[role] = set()
            permissions = self._permissions_by_role[role]
            for other_role, other_permissions in self._permissions_by_role.items():
                if other_role == role:
                    continue
                if other_permissions < permissions:
                    self._sub_roles_by_role[role].add(other_role)
        return self._sub_roles_by_role[role]

    def get_root_permissions(self) -> set[Permission]:
        """
        Get all possible permissions.
        """
        permissions: set[Permission] = self.app.domain.get_permissions_for_domain(frozen=False)  # type: ignore[assignment]
        return permissions

    @abc.abstractmethod
    def retrieve_user_roles(self, user: User) -> set[Hashable]:
        """
        Implement this method to retrieve the roles of a user.
        """
        raise NotImplementedError

    def retrieve_user_is_non_rbac_authorized(self, cmd: Command) -> bool:
        """
        Overriding this method provides a mechanism for a user to exist but
        (currently) not being authorized to execute the command in question,
        regardless of whether the user has the required permission through their roles.
        This is one way to implement non role-based access control.
        """
        return True

    def retrieve_user_is_root(self, user: User) -> bool:
        """
        Overriding this method provides a mechanism for a root user to exist, which is
        always allowed to execute any command.
        """
        return False

    def retrieve_user_permissions(self, user: User) -> set[Permission]:
        """
        Retrieve all permissions that a user has.
        """
        return set.union(
            set(),
            *[self._permissions_by_role[x] for x in self.retrieve_user_roles(user)],
        )

    def retrieve_user_has_all_rbac_permissions(self, user: User) -> bool:
        """
        Retrieve whether a user has all permissions that are subject to RBAC.
        """
        return (
            len(
                self.app.domain.permissions
                - self.retrieve_user_permissions(user)
                - self._permissions_without_rbac
            )
            == 0
        )

    def retrieve_user_has_more_permissions(
        self, user: User, tgt_user_or_roles: User | set[Hashable]
    ) -> bool:
        """
        Retrieve whether a user has more permissions than a target user or a set of
        roles. This is useful for determining whether a user can perform an action on
        another user.
        """
        if isinstance(tgt_user_or_roles, User):
            permissions = self.retrieve_user_permissions(tgt_user_or_roles)
        else:
            permissions = set.union(
                set(),
                *[self._permissions_by_role[x] for x in tgt_user_or_roles],
            )
        return not self.retrieve_user_permissions(user).issubset(permissions)

    def register_rbac_policies(self, **kwargs: Any) -> None:
        """
        Register an RBAC policy for every command that is subject to RBAC. In case of
        dynamic roles, i.e. that can be changed at runtime, these policies remain
        up-to-date as long as the updates are submitted to this service.

        The default implementation of the policy functions can be overridden by passing
        them as keyword arguments. This is useful for testing or for adding additional
        functionality to the policy functions. The following functions can be overridden:
        - get_permission_for_command
        - get_permission_has_rbac
        - get_roles_by_permission
        - retrieve_user_roles
        - retrieve_user_is_non_rbac_authorized
        - retrieve_user_is_root
        """
        # Create arguments for the RBAC policy constructor
        get_permission_for_command = kwargs.get(
            "get_permission_for_command",
            lambda x: self.app.domain.get_permission_for_command_instance(x),
        )
        get_permission_has_rbac = kwargs.get(
            "get_permission_has_rbac", lambda x: x not in self._permissions_without_rbac
        )
        get_roles_by_permission = kwargs.get(
            "get_roles_by_permission", lambda x: self._roles_by_permission[x]
        )
        retrieve_user_roles = kwargs.get(
            "retrieve_user_roles", lambda x: self.retrieve_user_roles(x)
        )
        retrieve_user_is_non_rbac_authorized = kwargs.get(
            "retrieve_user_is_non_rbac_authorized",
            lambda x: self.retrieve_user_is_non_rbac_authorized(x),
        )
        retrieve_user_is_root = kwargs.get(
            "retrieve_user_is_root", lambda x: self.retrieve_user_is_root(x)
        )
        # Create a single RBAC policy with the provided functions
        rbac_policy = RbacPolicy(
            get_permission_for_command,
            get_permission_has_rbac,
            get_roles_by_permission,
            retrieve_user_roles,
            retrieve_user_is_non_rbac_authorized=retrieve_user_is_non_rbac_authorized,
            retrieve_user_is_root=retrieve_user_is_root,
        )
        # Register the RBAC policy for all command classes that are subject to RBAC
        for command_class in self.get_command_classes_with_rbac():
            self.app.register_policy(command_class, rbac_policy, EventTiming.BEFORE)

    @staticmethod
    def expand_hierarchical_role_permissions(
        role_hierarchy: dict[Hashable, set[Hashable]],
        role_permission_sets: dict[
            Hashable, set[tuple[type[Command], PermissionTypeSet]]
        ],
        verify_redundant_permissions: bool = True,
    ) -> dict[Hashable, set[tuple[type[Command], PermissionType]]]:
        """
        Add the permissions of all sub-roles in a role hierarchy to the current role.
        This is useful to describe all permissions only for the lowest roles in the
        hierarchy, avoiding code duplication and risk of inconsistencies.

        Permissions are given as a tuple[Command, PermissionTypeSet], so that this
        function can be called before the actual permissions are created by the domain.

        The role hierarchy is a dictionary where the keys are the roles and the values
        are all the roles that are below the key role in the hierarchy, i.e. not only
        the direct sub-roles.

        If verify_redundant_permissions is True, the function will raise an error if
        there are any redundant permissions in the role hierarchy, whereby a parent
        role has the same permission as one of its sub-roles. This can be help to
        ensure that the role hierarchy is indeed correctly defined.
        """

        def _get_permissions(
            role_permission_sets: set[tuple[type[Command], PermissionTypeSet]],
        ) -> set[tuple[type[Command], PermissionType]]:
            permissions = set()
            for command_class, permission_type_set in role_permission_sets:
                permissions.update(
                    {(command_class, x) for x in permission_type_set.value}
                )
            return permissions

        role_permissions_map: dict[
            Hashable, set[tuple[type[Command], PermissionType]]
        ] = {}
        for role in role_hierarchy:
            role_permissions_map[role] = _get_permissions(
                role_permission_sets.get(role, set())
            )
            if verify_redundant_permissions:
                orig_role_permissions = role_permissions_map[role].copy()
            else:
                orig_role_permissions = set()
            for sub_role in role_hierarchy.get(role, set()):
                # Add the permissions of the sub-role to the role
                sub_role_permissions = _get_permissions(
                    role_permission_sets.get(sub_role, set())
                )
                role_permissions_map[role].update(sub_role_permissions)
                if not verify_redundant_permissions:
                    continue
                # Check if any sub-role permissions are also in the role's unique
                # permissions
                redundant_permissions = sub_role_permissions.intersection(
                    orig_role_permissions
                )
                if not redundant_permissions:
                    continue
                # Some redundant permissions -> summarize them by command class and
                # raise an error
                command_classes = sorted(
                    {x[0] for x in redundant_permissions},
                    key=lambda x: x.__name__,
                )
                redundant_permission_sets = []
                for command_class in command_classes:
                    redundant_permission_sets.append(
                        PermissionTypeSet(
                            frozenset(
                                {
                                    x[1]
                                    for x in redundant_permissions
                                    if x[0] == command_class
                                }
                            )
                        )
                    )
                duplicate_permissions_str = ", ".join(
                    [
                        f"{x.__name__}.{y.name}"
                        for x, y in zip(command_classes, redundant_permission_sets)
                    ]
                )
                role_str = role.value if isinstance(role, Enum) else str(role)
                sub_role_str = (
                    sub_role.value if isinstance(sub_role, Enum) else str(sub_role)
                )
                raise exc.InitializationServiceError(
                    f"Duplicate permissions in role hierarchy for role {role_str} and sub-role {sub_role_str}: {duplicate_permissions_str}"
                )
        return role_permissions_map



from enum import Enum


class SortOrder(Enum):
    ASCENDING = "ASCENDING"
    DESCENDING = "DESCENDING"


class PermissionType(Enum):
    CREATE = "CREATE"
    READ = "READ"
    UPDATE = "UPDATE"
    DELETE = "DELETE"
    EXECUTE = "EXECUTE"


class PermissionTypeSet(Enum):
    E = frozenset({PermissionType.EXECUTE})
    CRUD = frozenset(
        {
            PermissionType.CREATE,
            PermissionType.READ,
            PermissionType.UPDATE,
            PermissionType.DELETE,
        }
    )
    CRU = frozenset(
        {
            PermissionType.CREATE,
            PermissionType.READ,
            PermissionType.UPDATE,
        }
    )  # Undeletable
    CRD = frozenset(
        {
            PermissionType.CREATE,
            PermissionType.READ,
            PermissionType.DELETE,
        }
    )  # Immutable and deletable
    CUD = frozenset(
        {
            PermissionType.CREATE,
            PermissionType.UPDATE,
            PermissionType.DELETE,
        }
    )
    RUD = frozenset(
        {
            PermissionType.READ,
            PermissionType.UPDATE,
            PermissionType.DELETE,
        }
    )
    CR = frozenset({PermissionType.CREATE, PermissionType.READ})  # Immutable
    CU = frozenset({PermissionType.CREATE, PermissionType.UPDATE})
    CD = frozenset({PermissionType.CREATE, PermissionType.DELETE})
    RU = frozenset({PermissionType.READ, PermissionType.UPDATE})
    RD = frozenset({PermissionType.READ, PermissionType.DELETE})
    UD = frozenset({PermissionType.UPDATE, PermissionType.DELETE})
    C = frozenset({PermissionType.CREATE})
    R = frozenset({PermissionType.READ})  # Read only
    U = frozenset({PermissionType.UPDATE})
    D = frozenset({PermissionType.DELETE})
    NONE = frozenset()


class CrudOperation(Enum):
    CREATE_ONE = "CREATE_ONE"
    CREATE_SOME = "CREATE_SOME"
    READ_ALL = "READ_ALL"
    READ_SOME = "READ_SOME"
    READ_ONE = "READ_ONE"
    UPDATE_ONE = "UPDATE_ONE"
    UPDATE_SOME = "UPDATE_SOME"
    UPSERT_ONE = "UPSERT_ONE"
    UPSERT_SOME = "UPSERT_SOME"
    DELETE_ONE = "DELETE_ONE"
    DELETE_SOME = "DELETE_SOME"
    DELETE_ALL = "DELETE_ALL"
    UNDELETE_ONE = "UNDELETE_ONE"
    UNDELETE_SOME = "UNDELETE_SOME"
    UNDELETE_ALL = "UNDELETE_ALL"
    RESTORE_ONE = "RESTORE_ONE"
    RESTORE_SOME = "RESTORE_SOME"
    RESTORE_ALL = "RESTORE_ALL"
    EXISTS_ONE = "EXISTS_ONE"
    EXISTS_SOME = "EXISTS_SOME"


class CrudOperationSet(Enum):
    CREATE = frozenset({CrudOperation.CREATE_ONE, CrudOperation.CREATE_SOME})
    READ = frozenset(
        {
            CrudOperation.READ_ALL,
            CrudOperation.READ_SOME,
            CrudOperation.READ_ONE,
        }
    )
    READ_NOT_ALL = frozenset(
        {
            CrudOperation.READ_SOME,
            CrudOperation.READ_ONE,
        }
    )
    READ_OR_EXISTS = frozenset(
        {
            CrudOperation.READ_ALL,
            CrudOperation.READ_SOME,
            CrudOperation.READ_ONE,
            CrudOperation.EXISTS_ONE,
            CrudOperation.EXISTS_SOME,
        }
    )
    UPDATE = frozenset({CrudOperation.UPDATE_ONE, CrudOperation.UPDATE_SOME})
    DELETE = frozenset(
        {
            CrudOperation.DELETE_ONE,
            CrudOperation.DELETE_SOME,
            CrudOperation.DELETE_ALL,
        }
    )
    EXISTS = frozenset({CrudOperation.EXISTS_ONE, CrudOperation.EXISTS_SOME})
    NON_READ = frozenset(
        {
            CrudOperation.CREATE_ONE,
            CrudOperation.CREATE_SOME,
            CrudOperation.EXISTS_ONE,
            CrudOperation.EXISTS_SOME,
            CrudOperation.UPDATE_ONE,
            CrudOperation.UPDATE_SOME,
            CrudOperation.UPSERT_ONE,
            CrudOperation.UPSERT_SOME,
            CrudOperation.DELETE_ONE,
            CrudOperation.DELETE_SOME,
            CrudOperation.DELETE_ALL,
        }
    )
    NON_READ_OR_EXISTS = frozenset(
        {
            CrudOperation.CREATE_ONE,
            CrudOperation.CREATE_SOME,
            CrudOperation.UPDATE_ONE,
            CrudOperation.UPDATE_SOME,
            CrudOperation.UPSERT_ONE,
            CrudOperation.UPSERT_SOME,
            CrudOperation.DELETE_ONE,
            CrudOperation.DELETE_SOME,
            CrudOperation.DELETE_ALL,
        }
    )
    WRITE = frozenset(
        {
            CrudOperation.CREATE_ONE,
            CrudOperation.CREATE_SOME,
            CrudOperation.UPDATE_ONE,
            CrudOperation.UPDATE_SOME,
            CrudOperation.UPSERT_ONE,
            CrudOperation.UPSERT_SOME,
        }
    )
    UNDELETE = frozenset(
        {
            CrudOperation.UNDELETE_ONE,
            CrudOperation.UNDELETE_SOME,
            CrudOperation.UNDELETE_ALL,
        }
    )
    RESTORE = frozenset(
        {
            CrudOperation.RESTORE_ONE,
            CrudOperation.RESTORE_SOME,
        }
    )
    WRITE_ONE = frozenset(
        {
            CrudOperation.CREATE_ONE,
            CrudOperation.UPDATE_ONE,
            CrudOperation.UPSERT_ONE,
        }
    )
    WRITE_SOME = frozenset(
        {
            CrudOperation.CREATE_SOME,
            CrudOperation.UPDATE_SOME,
            CrudOperation.UPSERT_SOME,
        }
    )
    NON_CREATE_ONE = frozenset(
        {
            CrudOperation.READ_ONE,
            CrudOperation.UPDATE_ONE,
            CrudOperation.DELETE_ONE,
            CrudOperation.EXISTS_ONE,
        }
    )
    NON_CREATE_SOME = frozenset(
        {
            CrudOperation.READ_SOME,
            CrudOperation.UPDATE_SOME,
            CrudOperation.DELETE_SOME,
            CrudOperation.EXISTS_SOME,
        }
    )
    ANY_ONE = frozenset(
        {
            CrudOperation.CREATE_ONE,
            CrudOperation.READ_ONE,
            CrudOperation.UPDATE_ONE,
            CrudOperation.DELETE_ONE,
            CrudOperation.EXISTS_ONE,
        }
    )
    ANY_ALL = frozenset(
        {
            CrudOperation.READ_ALL,
            CrudOperation.DELETE_ALL,
        }
    )


class HttpProtocol(Enum):
    HTTP = "HTTP"
    HTTPS = "HTTPS"


class CrudEndpointType(Enum):
    POST_ONE = "POST_ONE"
    POST_SOME = "POST_SOME"
    GET_ALL = "GET_ALL"
    GET_SOME = "GET_SOME"
    GET_ONE = "GET_ONE"
    PUT_ONE = "PUT_ONE"
    PUT_SOME = "PUT_SOME"
    DELETE_ONE = "DELETE_ONE"
    DELETE_SOME = "DELETE_SOME"
    DELETE_ALL = "DELETE_ALL"
    POST_QUERY = "POST_QUERY"
    POST_QUERY_IDS = "POST_QUERY_IDS"


class IsolationLevel(Enum):
    READ_UNCOMMITED = "READ_UNCOMMITED"
    READ_COMMITED = "READ_COMMITED"
    REPEATABLE_READ = "REPEATABLE_READ"
    SERIALIZABLE = "SERIALIZABLE"


class LogLevel(Enum):
    FATAL = "FATAL"
    ERROR = "ERROR"
    WARN = "WARN"
    INFO = "INFO"
    DEBUG = "DEBUG"
    TRACE = "TRACE"


class EventTiming(Enum):
    BEFORE = "BEFORE"
    DURING = "DURING"
    AFTER = "AFTER"


class AuthProtocol(Enum):
    NONE = "NONE"
    OAUTH2 = "OAUTH2"
    OIDC = "OIDC"


class OAuthFlow(Enum):
    AUTHORIZATION_CODE = "AUTHORIZATION_CODE"
    CLIENT_CREDENTIALS = "CLIENT_CREDENTIALS"
    RESOURCE_OWNER = "RESOURCE_OWNER"
    HYBRID = "HYBRID"
    DEVICE_AUTHORIZATION = "DEVICE_AUTHORIZATION"
    PKCE = "PKCE"


class HttpMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    HEAD = "HEAD"
    CONNECT = "CONNECT"
    OPTIONS = "OPTIONS"
    TRACE = "TRACE"


class StringCasing(Enum):
    SNAKE_CASE = "SNAKE_CASE"
    CAMEL_CASE = "CAMEL_CASE"
    PASCAL_CASE = "PASCAL_CASE"
    KEBAB_CASE = "KEBAB_CASE"


class FieldType(Enum):
    """
    This enumeration is used to categorize the different types of fields that can be
    present in a model as well as in its persisted version.

    Attributes
    ----------
    ID : Enum
        Represents an identifier field in the model.
    LINK : Enum
        Represents a link field in the model.
    VALUE : Enum
        Represents a value field in the model.
    BACK_POPULATE : Enum
        Represents a back-populate field in the model.
    SERVICE_METADATA : Enum
        Represents a metadata field generated by the service.
    DB_METADATA : Enum
        Represents a metadata field generated by the repository.
    """

    ID = "ID"
    LINK = "LINK"
    VALUE = "VALUE"
    COMPUTED = "COMPUTED"
    RELATIONSHIP = "RELATIONSHIP"
    SERVICE_METADATA = "SERVICE_METADATA"
    DB_METADATA = "DB_METADATA"


class FieldTypeSet(Enum):
    """
    Different sets of field types that can be used to categorize the fields in a model.
    The value is a tuple instead of a frozenset to guarantee the order of the elements.
    """

    ID = tuple([FieldType.ID])
    LINK = tuple([FieldType.LINK])
    VALUE = tuple([FieldType.VALUE])
    RELATIONSHIP = tuple([FieldType.RELATIONSHIP])
    COMPUTED = tuple([FieldType.COMPUTED])
    SERVICE_METADATA = tuple([FieldType.SERVICE_METADATA])
    DB_METADATA = tuple([FieldType.DB_METADATA])
    DATA = tuple([FieldType.LINK, FieldType.VALUE])
    METADATA = tuple([FieldType.SERVICE_METADATA, FieldType.DB_METADATA])
    MODEL_DB_COMMON = tuple([FieldType.ID, FieldType.LINK, FieldType.VALUE])
    MODEL = tuple(
        [
            FieldType.ID,
            FieldType.LINK,
            FieldType.VALUE,
            FieldType.RELATIONSHIP,
            FieldType.COMPUTED,
        ]
    )
    MODEL_ONLY = tuple([FieldType.RELATIONSHIP, FieldType.COMPUTED])
    DB_MODEL_ONLY = tuple([FieldType.SERVICE_METADATA, FieldType.DB_METADATA])
    DB_MODEL = tuple(
        [
            FieldType.ID,
            FieldType.LINK,
            FieldType.VALUE,
            FieldType.SERVICE_METADATA,
            FieldType.DB_METADATA,
        ]
    )



from collections.abc import Callable
from types import NoneType, UnionType
from typing import Any, Union

from typing_extensions import Annotated, get_args, get_origin

from gen_epix.fastapp.domain.key import Key
from gen_epix.fastapp.domain.link import Link


def create_keys(keys: dict[int, Key | str | tuple | Callable]) -> dict[int, Key]:
    retval = {}
    for x, y in keys.items():
        if isinstance(y, Key):
            retval[x] = y
        else:
            retval[x] = Key(y)
    return retval


def create_links(
    links: dict[int, Link | tuple[str, type, str | None]],
) -> dict[int, Link]:
    retval = {}
    for x, y in links.items():
        if isinstance(y, Link):
            retval[x] = y
        else:
            retval[x] = Link(
                link_field_name=y[0],
                link_model_class=y[1],
                relationship_field_name=y[2],
            )
    return retval


def get_type_from_annotation(
    annotation: type[Any] | None,
) -> type:
    """
    Adapted from https://github.com/fastapi/sqlmodel v0.0.24
    """
    # Resolve Optional fields
    if annotation is None:
        raise ValueError("Missing field type")
    origin = get_origin(annotation)
    if origin is None:
        return annotation
    elif origin is Annotated:
        return get_type_from_annotation(get_args(annotation)[0])
    if origin is UnionType or origin is Union:
        bases = get_args(annotation)
        if len(bases) > 2:
            raise ValueError("Field type is a non-optional union")
        # Non optional unions are not allowed
        assert len(bases) == 2
        if bases[0] is not NoneType and bases[1] is not NoneType:
            raise ValueError("Field type is a non-optional union")
        # Optional unions are allowed
        use_type = bases[0] if bases[0] is not NoneType else bases[1]
        return get_type_from_annotation(use_type)
    return origin  # type:ignore[no-any-return]



from typing import Any

from gen_epix.fastapp import exc, model
from gen_epix.fastapp.enum import EventTiming


class PolicyDecisionPoint:
    """
    Policy Decision Point (PDP). This is the central point where policies are
    registered and applied. They are executed in order of registration.

    Policies must be registered for a particular command and timing (BEFORE, DURING
    or AFTER) of execution:
    - BEFORE: raise UnauthorizedAuthorityError if any policy denies the command.
    - DURING: add policies to command so that they can be used during command execution.
    - AFTER: filter the return value with each policy and return it.
    """

    def __init__(self) -> None:
        self._policies: dict[
            type[model.Command], dict[EventTiming, list[model.Policy]]
        ] = {}

    def register_policy(
        self,
        command_class: type[model.Command],
        policy: model.Policy,
        timing: EventTiming = EventTiming.BEFORE,
    ) -> None:
        """
        Register a policy for a command class and timing before, during or after command execution.
        """
        if command_class not in self._policies:
            self._policies[command_class] = {}
        if timing not in self._policies[command_class]:
            self._policies[command_class][timing] = []
        if policy in self._policies[command_class][timing]:
            raise exc.InitializationServiceError(
                f"Policy {policy} already registered for command class {command_class} and timing {timing}"
            )
        self._policies[command_class][timing].append(policy)

    def unregister_policy(
        self,
        command_class: type[model.Command],
        policy: model.Policy,
        timing: EventTiming | None = None,
    ) -> None:
        """
        Unregister a policy for a command class and timing (all timings if None).
        """
        if command_class not in self._policies:
            raise exc.InitializationServiceError(
                f"No policies registered for command class {command_class}"
            )
        if timing is not None:
            if timing not in self._policies[command_class]:
                raise exc.InitializationServiceError(
                    f"No policies registered for command class {command_class} and timing {timing}"
                )
            if policy not in self._policies[command_class][timing]:
                raise exc.InitializationServiceError(
                    f"Policy {policy} not registered for command class {command_class} and timing {timing}"
                )
            self._policies[command_class][timing].remove(policy)
        else:
            has_policy = False
            for timing in self._policies[command_class]:
                if policy in self._policies[command_class][timing]:
                    has_policy = True
                    self._policies[command_class][timing].remove(policy)
            if not has_policy:
                raise exc.InitializationServiceError(
                    f"Policy {policy} not registered for command class {command_class}, any timing"
                )

    def get_policies(
        self, command_class: type[model.Command], timing: EventTiming
    ) -> list[model.Policy]:
        """
        Get all policies registered for a command class and timing. The list is a copy.
        """
        return list(self._policies.get(command_class, {}).get(timing, []))

    def apply(
        self, cmd: model.Command, timing: EventTiming, retval: Any | None = None
    ) -> Any | None:
        """
        Apply policies for a command class and timing.
        In case of BEFORE, raise unauthorized error if any policy denies the command.
        In case of DURING, add policies to command so that they can be used during command execution.
        In case of AFTER, filter the return value with each policy and return it.
        """
        policies = self.get_policies(type(cmd), timing)
        if not policies:
            return retval if timing == EventTiming.AFTER else None
        if timing == EventTiming.BEFORE:
            for policy in policies:
                if not policy.is_allowed(cmd):
                    raise policy.get_is_denied_exception()(
                        f"Policy {policy.__class__.__name__} denied {cmd.__class__.__name__} command {cmd}"
                    )
            return None
        elif timing == EventTiming.DURING:
            # Add policies to command so that they can be used during command execution
            cmd._policies.extend(policies)
            return None
        elif timing == EventTiming.AFTER:
            # Execute policies that may alter the return value
            for policy in policies:
                retval = policy.filter(cmd, retval)
            return retval



# pylint: disable=useless-import-alias
from gen_epix.casedb.services.abac import AbacService as AbacService
from gen_epix.casedb.services.case import CaseService as CaseService
from gen_epix.casedb.services.geo import GeoService as GeoService
from gen_epix.casedb.services.ontology import OntologyService as OntologyService
from gen_epix.casedb.services.seqdb import SeqdbRemoteApp as SeqdbRemoteApp
from gen_epix.casedb.services.seqdb import SeqdbService as SeqdbService
from gen_epix.casedb.services.subject import SubjectService as SubjectService
from gen_epix.commondb.services import AuthService as AuthService
from gen_epix.commondb.services import OrganizationService as OrganizationService
from gen_epix.commondb.services import SystemService as SystemService
from gen_epix.commondb.services import UserManager as UserManager
from gen_epix.commondb.services.rbac import RbacService as RbacService
